# Copyright (C) 2007, 2011, One Laptop Per Child
# Copyright (C) 2014, Ignacio Rodriguez
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
from gettext import gettext as _

from gi.repository import GObject
from gi.repository import Gio
from gi.repository import GLib
from gi.repository import Gtk
from gi.repository import Gdk
import pickle
import xapian
import json
import tempfile
import shutil

from sugar3.graphics.radiotoolbutton import RadioToolButton
from sugar3.graphics.palette import Palette
from sugar3.graphics import style
from sugar3 import env
from sugar3 import profile

from jarabe.journal import model
from jarabe.journal.misc import get_mount_icon_name
from jarabe.journal.misc import get_mount_color
from jarabe.view.palettes import VolumePalette


_JOURNAL_0_METADATA_DIR = '.olpc.store'


def _get_id(document):
    """Get the ID for the document in the xapian database."""
    tl = document.termlist()
    try:
        term = tl.skip_to('Q').term
        if len(term) == 0 or term[0] != 'Q':
            return None
        return term[1:]
    except StopIteration:
        return None


def _convert_entries(root):
    """Convert entries written by the datastore version 0.

    The metadata and the preview will be written using the new
    scheme for writing Journal entries to removable storage
    devices.

    - entries that do not have an associated file are not
    converted.
    - if an entry has no title we set it to Untitled and rename
    the file accordingly, taking care of creating a unique
    filename

    """
    try:
        database = xapian.Database(os.path.join(root, _JOURNAL_0_METADATA_DIR,
                                                'index'))
    except xapian.DatabaseError:
        logging.exception('Convert DS-0 Journal entries: error reading db: %s',
                          os.path.join(root, _JOURNAL_0_METADATA_DIR, 'index'))
        return

    metadata_dir_path = os.path.join(root, model.JOURNAL_METADATA_DIR)
    if not os.path.exists(metadata_dir_path):
        try:
            os.mkdir(metadata_dir_path)
        except EnvironmentError:
            logging.error('Convert DS-0 Journal entries: '
                          'error creating the Journal metadata directory.')
            return

    for posting_item in database.postlist(''):
        try:
            document = database.get_document(posting_item.docid)
        except xapian.DocNotFoundError as e:
            logging.debug('Convert DS-0 Journal entries: error getting '
                          'document %s: %s', posting_item.docid, e)
            continue
        _convert_entry(root, document)


def _convert_entry(root, document):
    try:
        metadata_loaded = pickle.loads(document.get_data())
    except pickle.PickleError as e:
        logging.debug('Convert DS-0 Journal entries: '
                      'error converting metadata: %s', e)
        return

    if not ('activity_id' in metadata_loaded and
            'mime_type' in metadata_loaded and
            'title' in metadata_loaded):
        return

    metadata = {}

    uid = _get_id(document)
    if uid is None:
        return

    for key, value in list(metadata_loaded.items()):
        metadata[str(key)] = str(value[0])

    if 'uid' not in metadata:
        metadata['uid'] = uid

    filename = metadata.pop('filename', None)
    if not filename:
        return
    if not os.path.exists(os.path.join(root, filename)):
        return

    if not metadata.get('title'):
        metadata['title'] = _('Untitled')
        fn = model.get_file_name(metadata['title'],
                                 metadata['mime_type'])
        new_filename = model.get_unique_file_name(root, fn)
        os.rename(os.path.join(root, filename),
                  os.path.join(root, new_filename))
        filename = new_filename

    preview_path = os.path.join(root, _JOURNAL_0_METADATA_DIR,
                                'preview', uid)
    if os.path.exists(preview_path):
        preview_fname = filename + '.preview'
        new_preview_path = os.path.join(root,
                                        model.JOURNAL_METADATA_DIR,
                                        preview_fname)
        if not os.path.exists(new_preview_path):
            shutil.copy(preview_path, new_preview_path)

    metadata_fname = filename + '.metadata'
    metadata_path = os.path.join(root, model.JOURNAL_METADATA_DIR,
                                 metadata_fname)
    if not os.path.exists(metadata_path):
        (fh, fn) = tempfile.mkstemp(dir=root)
        os.write(fh, json.dumps(metadata))
        os.close(fh)
        os.rename(fn, metadata_path)

        logging.debug('Convert DS-0 Journal entries: entry converted: '
                      'file=%s metadata=%s',
                      os.path.join(root, filename), metadata)


class VolumesToolbar(Gtk.Toolbar):
    __gtype_name__ = 'VolumesToolbar'

    __gsignals__ = {
        'volume-changed': (GObject.SignalFlags.RUN_FIRST, None,
                           ([str])),
        'volume-error': (GObject.SignalFlags.RUN_FIRST, None,
                         ([str, str])),
    }

    def __init__(self):
        Gtk.Toolbar.__init__(self)
        self._mount_added_hid = None
        self._mount_removed_hid = None

        button = JournalButton()
        button.connect('toggled', self._button_toggled_cb)
        self.insert(button, 0)
        button.show()
        self._volume_buttons = [button]

        self.connect('destroy', self.__destroy_cb)

        GLib.idle_add(self._set_up_volumes)

    def __destroy_cb(self, widget):
        volume_monitor = Gio.VolumeMonitor.get()
        volume_monitor.disconnect(self._mount_added_hid)
        volume_monitor.disconnect(self._mount_removed_hid)

    def _set_up_volumes(self):
        self._set_up_documents_button()

        volume_monitor = Gio.VolumeMonitor.get()
        self._mount_added_hid = volume_monitor.connect('mount-added',
                                                       self.__mount_added_cb)
        self._mount_removed_hid = volume_monitor.connect(
            'mount-removed',
            self.__mount_removed_cb)

        for mount in volume_monitor.get_mounts():
            self._add_button(mount)

    def _set_up_documents_button(self):
        documents_path = model.get_documents_path()
        if documents_path is not None:
            button = DocumentsButton(documents_path)
            button.props.group = self._volume_buttons[0]
            button.set_palette(Palette(_('Documents')))
            button.connect('toggled', self._button_toggled_cb)
            button.show()

            position = self.get_item_index(self._volume_buttons[-1]) + 1
            self.insert(button, position)
            self._volume_buttons.append(button)
            self.show()

    def __mount_added_cb(self, volume_monitor, mount):
        self._add_button(mount)

    def __mount_removed_cb(self, volume_monitor, mount):
        self._remove_button(mount)

    def _add_button(self, mount):
        logging.debug('VolumeToolbar._add_button: %r', mount.get_name())

        if os.path.exists(os.path.join(mount.get_root().get_path(),
                                       _JOURNAL_0_METADATA_DIR)):
            logging.debug('Convert DS-0 Journal entries: starting conversion')
            GLib.idle_add(_convert_entries, mount.get_root().get_path())

        button = VolumeButton(mount)
        button.props.group = self._volume_buttons[0]
        button.connect('toggled', self._button_toggled_cb)
        button.connect('volume-error', self.__volume_error_cb)
        position = self.get_item_index(self._volume_buttons[-1]) + 1
        self.insert(button, position)
        button.show()

        self._volume_buttons.append(button)

        if len(self.get_children()) > 1:
            self.show()

    def __volume_error_cb(self, button, strerror, severity):
        self.emit('volume-error', strerror, severity)

    def _button_toggled_cb(self, button):
        if button.props.active:
            self.emit('volume-changed', button.mount_point)

    def _get_button_for_mount(self, mount):
        mount_point = mount.get_root().get_path()
        for button in self.get_children():
            if button.mount_point == mount_point:
                return button
        logging.error('Couldnt find button with mount_point %r', mount_point)
        return None

    def _remove_button(self, mount):
        button = self._get_button_for_mount(mount)
        self._volume_buttons.remove(button)
        self.remove(button)
        self.get_children()[0].props.active = True

        if len(self.get_children()) < 2:
            self.hide()

    def set_active_volume(self, mount):
        button = self._get_button_for_mount(mount)
        button.props.active = True


class BaseButton(RadioToolButton):
    __gsignals__ = {
        'volume-error': (GObject.SignalFlags.RUN_FIRST, None,
                         ([str, str])),
    }

    def __init__(self, mount_point):
        RadioToolButton.__init__(self)

        self.mount_point = mount_point

        self.drag_dest_set(Gtk.DestDefaults.ALL,
                           [Gtk.TargetEntry.new('journal-object-id', 0, 0)],
                           Gdk.DragAction.COPY)
        self.connect('drag-data-received', self._drag_data_received_cb)

    def _drag_data_received_cb(self, widget, drag_context, x, y,
                               selection_data, info, timestamp):
        object_id = selection_data.get_data()
        metadata = model.get(object_id)
        file_path = model.get_file(metadata['uid'])
        if not file_path or not os.path.exists(file_path):
            logging.warn('Entries without a file cannot be copied.')
            self.emit('volume-error',
                      _('Entries without a file cannot be copied.'),
                      _('Warning'))
            return

        try:
            model.copy(metadata, self.mount_point)
        except IOError as e:
            logging.exception('Error while copying the entry. %s', e.strerror)
            self.emit('volume-error',
                      _('Error while copying the entry. %s') % e.strerror,
                      _('Error'))


class VolumeButton(BaseButton):

    def __init__(self, mount):
        self._mount = mount
        mount_point = mount.get_root().get_path()
        BaseButton.__init__(self, mount_point)

        self.props.icon_name = get_mount_icon_name(mount,
                                                   Gtk.IconSize.LARGE_TOOLBAR)
        # TODO: retrieve the colors from the owner of the device
        self.props.xo_color = get_mount_color(self._mount)

    def create_palette(self):
        palette = VolumePalette(self._mount)
        # palette.props.invoker = FrameWidgetInvoker(self)
        # palette.set_group_id('frame')
        return palette


class JournalButton(BaseButton):

    def __init__(self):
        BaseButton.__init__(self, mount_point='/')

        self.props.icon_name = 'activity-journal'
        self.props.xo_color = profile.get_color()

    def create_palette(self):
        palette = JournalButtonPalette(self)
        return palette


class JournalButtonPalette(Palette):

    def __init__(self, mount):
        Palette.__init__(self, _('Journal'))

        grid = Gtk.Grid(orientation=Gtk.Orientation.VERTICAL,
                        margin=style.DEFAULT_SPACING,
                        row_spacing=style.DEFAULT_SPACING)
        self.set_content(grid)
        grid.show()

        self._progress_bar = Gtk.ProgressBar()
        grid.add(self._progress_bar)
        self._progress_bar.show()

        self._free_space_label = Gtk.Label()
        self._free_space_label.set_alignment(0.5, 0.5)
        grid.add(self._free_space_label)
        self._free_space_label.show()

        self.connect('popup', self.__popup_cb)

    def __popup_cb(self, palette):
        stat = os.statvfs(env.get_profile_path())
        free_space = stat[0] * stat[4]
        total_space = stat[0] * stat[2]

        fraction = (total_space - free_space) / float(total_space)
        self._progress_bar.props.fraction = fraction
        self._free_space_label.props.label = _('%(free_space)d MiB Free') % \
            {'free_space': free_space / (1024 * 1024)}


class DocumentsButton(BaseButton):

    def __init__(self, documents_path):
        BaseButton.__init__(self, mount_point=documents_path)

        self.props.icon_name = 'user-documents'
        self.props.xo_color = profile.get_color()



# Copyright (C) 2006-2007 Red Hat, Inc.
# Copyright (C) 2010 Collabora Ltd. <http://www.collabora.co.uk/>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging

from gi.repository import GObject
import dbus

from gi.repository import TelepathyGLib
CONNECTION = TelepathyGLib.IFACE_CONNECTION
CONNECTION_STATUS_CONNECTED = TelepathyGLib.ConnectionStatus.CONNECTED

from sugar3 import profile

from jarabe.util.telepathy import connection_watcher


CONNECTION_INTERFACE_BUDDY_INFO = 'org.laptop.Telepathy.BuddyInfo'

_owner_instance = None


class BaseBuddyModel(GObject.GObject):
    __gtype_name__ = 'SugarBaseBuddyModel'

    def __init__(self, **kwargs):
        self._key = None
        self._nick = None
        self._color = None
        self._tags = None
        self._current_activity = None

        GObject.GObject.__init__(self, **kwargs)

    def get_nick(self):
        return self._nick

    def set_nick(self, nick):
        self._nick = nick

    nick = GObject.Property(type=object, getter=get_nick, setter=set_nick)

    def get_key(self):
        return self._key

    def set_key(self, key):
        if isinstance(key, bytes):
            key = key.decode('utf-8')

        self._key = key

    key = GObject.Property(type=object, getter=get_key, setter=set_key)

    def get_color(self):
        return self._color

    def set_color(self, color):
        self._color = color

    color = GObject.Property(type=object, getter=get_color, setter=set_color)

    def get_tags(self):
        return self._tags

    tags = GObject.Property(type=object, getter=get_tags)

    def get_current_activity(self):
        return self._current_activity

    def set_current_activity(self, current_activity):
        if self._current_activity != current_activity:
            self._current_activity = current_activity
            self.notify('current-activity')

    current_activity = GObject.Property(type=object,
                                        getter=get_current_activity,
                                        setter=set_current_activity)

    def is_owner(self):
        raise NotImplementedError


class OwnerBuddyModel(BaseBuddyModel):
    __gtype_name__ = 'SugarOwnerBuddyModel'

    def __init__(self):
        BaseBuddyModel.__init__(self)

        self.props.nick = profile.get_nick_name()
        self.props.color = profile.get_color()

        self.props.key = profile.get_profile().pubkey

        self.connect('notify::nick', self.__property_changed_cb)
        self.connect('notify::color', self.__property_changed_cb)

        bus = dbus.SessionBus()
        bus.add_signal_receiver(
            self.__name_owner_changed_cb,
            signal_name='NameOwnerChanged',
            dbus_interface='org.freedesktop.DBus')

        bus_object = bus.get_object(dbus.BUS_DAEMON_NAME, dbus.BUS_DAEMON_PATH)
        for service in bus_object.ListNames(
                dbus_interface=dbus.BUS_DAEMON_IFACE):
            if service.startswith(CONNECTION + '.'):
                path = '/%s' % service.replace('.', '/')
                conn_proxy = bus.get_object(service, path)
                self._prepare_conn(path, conn_proxy)

    def _prepare_conn(self, object_path, conn_proxy):
        self.connection = {}
        self.object_path = object_path
        self.conn_proxy = conn_proxy
        self.conn_ready = False
        self.connection[CONNECTION] = \
            dbus.Interface(self.conn_proxy, CONNECTION)
        self.connection[CONNECTION].GetInterfaces(
            reply_handler=self.__conn_get_interfaces_reply_cb,
            error_handler=self.__error_handler_cb)

    def __conn_get_interfaces_reply_cb(self, interfaces):
        for interface in interfaces:
            self.connection[interface] = dbus.Interface(
                self.conn_proxy, interface)
        self.conn_ready = True
        self.__connection_ready_cb(self.connection)

    def __connection_ready_cb(self, connection):
        if not self.conn_ready:
            return

        self._sync_properties_on_connection(connection)

    def __name_owner_changed_cb(self, name, old, new):
        if name.startswith(CONNECTION + '.') and not old and new:
            path = '/' + name.replace('.', '/')
            self.conn_proxy = dbus.Bus().get_object(name, path)
            self._prepare_conn(path, self.conn_proxy)

    def __property_changed_cb(self, buddy, pspec):
        self._sync_properties()

    def _sync_properties(self):
        conn_watcher = connection_watcher.get_instance()
        for connection in conn_watcher.get_connections():
            self._sync_properties_on_connection(connection)

    def _sync_properties_on_connection(self, connection):
        if CONNECTION_INTERFACE_BUDDY_INFO in connection:
            properties = {}
            if self.props.key is not None:
                properties['key'] = dbus.ByteArray(
                    self.props.key.encode('utf-8'))
            if self.props.color is not None:
                properties['color'] = self.props.color.to_string()

            logging.debug('calling SetProperties with %r', properties)
            connection[CONNECTION_INTERFACE_BUDDY_INFO].SetProperties(
                properties,
                reply_handler=self.__set_properties_cb,
                error_handler=self.__error_handler_cb)

    def __set_properties_cb(self):
        logging.debug('__set_properties_cb')

    def __error_handler_cb(self, error):
        raise RuntimeError(error)

    def is_owner(self):
        return True


def get_owner_instance():
    global _owner_instance
    if _owner_instance is None:
        _owner_instance = OwnerBuddyModel()
    return _owner_instance


class BuddyModel(BaseBuddyModel):
    __gtype_name__ = 'SugarBuddyModel'

    def __init__(self, **kwargs):

        self._account = None
        self._contact_id = None
        self._handle = None

        BaseBuddyModel.__init__(self, **kwargs)

    def is_owner(self):
        return False

    def get_account(self):
        return self._account

    def set_account(self, account):
        self._account = account

    account = GObject.Property(type=object, getter=get_account,
                               setter=set_account)

    def get_contact_id(self):
        return self._contact_id

    def set_contact_id(self, contact_id):
        self._contact_id = contact_id

    contact_id = GObject.Property(type=object, getter=get_contact_id,
                                  setter=set_contact_id)

    def get_handle(self):
        return self._handle

    def set_handle(self, handle):
        self._handle = handle

    handle = GObject.Property(type=object, getter=get_handle,
                              setter=set_handle)



# Copyright (C) 2007, Eduardo Silva <edsiper@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gdk

from sugar3.graphics import style
from sugar3.graphics.palette import WidgetInvoker


def _get_screen_area():
    frame_thickness = style.GRID_CELL_SIZE

    screen_area = Gdk.Rectangle()
    screen_area.x = screen_area.y = frame_thickness
    screen_area.width = Gdk.Screen.width() - frame_thickness
    screen_area.height = Gdk.Screen.height() - frame_thickness

    return screen_area


class FrameWidgetInvoker(WidgetInvoker):

    def __init__(self, widget):
        WidgetInvoker.__init__(self, widget, widget.get_child())

        self._position_hint = self.ANCHORED
        self._screen_area = _get_screen_area()



# Copyright (C) 2014 Sam Parkinson
# Copyright (C) 2015 Martin Abente Lahaye
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gettext import gettext as _
import math

from gi.repository import GLib
from gi.repository import Gtk

from sugar3 import profile
from sugar3.graphics import style
from sugar3.graphics.icon import Icon
from sugar3.graphics.tray import TrayIcon
from sugar3.graphics.palette import Palette
from sugar3.graphics.palettemenu import PaletteMenuBox
from sugar3.graphics.palettemenu import PaletteMenuItem
from sugar3.graphics.palettemenu import PaletteMenuItemSeparator
from sugar3.graphics.xocolor import XoColor

from jarabe.frame.frameinvoker import FrameWidgetInvoker
from jarabe.model import brightness
from jarabe.model.screenshot import take_screenshot
from jarabe import frame


class DeviceView(TrayIcon):

    FRAME_POSITION_RELATIVE = 160

    def __init__(self, label):
        self._color = profile.get_color()
        self._label = label

        TrayIcon.__init__(self, icon_name='brightness-100',
                          xo_color=self._color)

        self.set_palette_invoker(FrameWidgetInvoker(self))
        self.palette_invoker.props.toggle_palette = True

        model = brightness.get_instance()
        if model.get_path():
            self._model = model
            self._model.changed_signal.connect(self.__brightness_changed_cb)
            self._update_output_info()

    def create_palette(self):
        palette = DisplayPalette()
        palette.set_group_id('frame')
        return palette

    def _update_output_info(self, value=None):
        if value is None:
            value = self._model.get_brightness()

        icon_number = math.ceil(
            float(value) * 3 / self._model.get_max_brightness()) * 33

        if icon_number == 99:
            icon_number = 100

        self.icon.props.icon_name = \
            'brightness-{:03d}'.format(int(icon_number))

    def __brightness_changed_cb(self, model, value):
        self._update_output_info(value)


class BrightnessManagerWidget(Gtk.VBox):

    TIMEOUT_DELAY = 10

    def __init__(self, text, icon_name):
        Gtk.VBox.__init__(self)
        self._progress_bar = None
        self._adjustment = None

        icon = Icon(pixel_size=style.SMALL_ICON_SIZE)
        icon.props.icon_name = icon_name
        icon.props.xo_color = XoColor('%s,%s' % (style.COLOR_WHITE.get_svg(),
                                      style.COLOR_BUTTON_GREY.get_svg()))
        icon.show()

        label = Gtk.Label(text)
        label.show()

        grid = Gtk.Grid()
        grid.set_column_spacing(style.DEFAULT_SPACING)
        grid.attach(icon, 0, 0, 1, 1)
        grid.attach(label, 1, 0, 1, 1)
        grid.show()

        alignment = Gtk.Alignment()
        alignment.set(0.5, 0, 0, 0)
        alignment.add(grid)
        alignment.show()
        self.add(alignment)

        alignment = Gtk.Alignment()
        alignment.set(0.5, 0, 0, 0)
        alignment.set_padding(0, 0, style.DEFAULT_SPACING,
                              style.DEFAULT_SPACING)

        self._model = brightness.get_instance()
        self._model_changed_hid = \
            self._model.changed_signal.connect(self.__brightness_changed_cb)

        # if sugar-backlight-helper finds the device
        if self._model.get_path():
            adjustment = Gtk.Adjustment(
                value=self._model.get_brightness(),
                lower=0,
                upper=self._model.get_max_brightness() + 1,
                step_incr=self._model.get_step_amount(),
                page_incr=self._model.get_step_amount(),
                page_size=self._model.get_step_amount())
            self._adjustment = adjustment

            self._adjustment_timeout_id = None
            self._adjustment_hid = \
                self._adjustment.connect('value-changed', self.__adjusted_cb)

            hscale = Gtk.HScale()
            hscale.props.draw_value = False
            hscale.set_adjustment(adjustment)
            hscale.set_digits(0)
            hscale.set_size_request(style.GRID_CELL_SIZE * 4, -1)
            alignment.add(hscale)
            hscale.show()
        else:
            self._progress_bar = Gtk.ProgressBar()
            self._progress_bar.set_size_request(
                style.zoom(style.GRID_CELL_SIZE * 4), -1)
            alignment.props.top_padding = style.DEFAULT_PADDING
            alignment.add(self._progress_bar)
            self._progress_bar.show()

        alignment.show()
        self.add(alignment)

    def __brightness_changed_cb(self, model, value):
        self.update(value)

    def update(self, value=None):
        if value is None:
            value = self._model.get_brightness()

        if self._adjustment:
            self._adjustment.handler_block(self._adjustment_hid)
            self._adjustment.props.value = value
            self._adjustment.handler_unblock(self._adjustment_hid)
        else:
            self._progress_bar.props.fraction = \
                float(value) / self._model.get_max_brightness()

    def __adjusted_cb(self, device, data=None):
        if self._adjustment_timeout_id is not None:
            GLib.source_remove(self._adjustment_timeout_id)
        self._adjustment_timeout_id = GLib.timeout_add(
            self.TIMEOUT_DELAY, self._adjust_brightness)

    def _adjust_brightness(self):
        self._model.handler_block(self._model_changed_hid)
        self._model.set_brightness(int(self._adjustment.props.value))
        self._model.handler_unblock(self._model_changed_hid)
        self._adjustment_timeout_id = None
        return False


class DisplayPalette(Palette):

    def __init__(self):
        Palette.__init__(self, label=_('My Display'))

        self._screenshot = PaletteMenuItem(_('Take a screenshot'))
        icon = Icon(icon_name='camera-external',
                    pixel_size=style.SMALL_ICON_SIZE)
        self._screenshot.set_image(icon)
        icon.show()
        self._screenshot.connect('activate', self.__screenshot_cb)
        self._screenshot.show()

        self._box = PaletteMenuBox()

        self._brightness_manager = None
        # only add this widget if device available
        if brightness.get_instance().get_path():
            self._add_brightness_manager()

        self._box.append_item(self._screenshot, 0, 0)
        self._box.show()

        self.set_content(self._box)
        self.connect('popup', self.__popup_cb)

    def _add_brightness_manager(self):
        self._brightness_manager = BrightnessManagerWidget(_('Brightness'),
                                                           'brightness-100')
        self._brightness_manager.show()

        separator = PaletteMenuItemSeparator()
        separator.show()

        self._box.append_item(self._brightness_manager, 0, 0)
        self._box.append_item(separator, 0, 0)

    def __popup_cb(self, palette):
        if self._brightness_manager is not None:
            self._brightness_manager.update()

    def __screenshot_cb(self, palette):
        frame_ = frame.get_view()
        frame_.hide()
        GLib.idle_add(self.__take_screenshot_cb, frame_)

    def __take_screenshot_cb(self, frame_):
        if frame_.is_visible():
            return True
        take_screenshot()
        frame_.show()
        return False


def setup(tray):
    tray.add_device(DeviceView(_('Display')))



# Copyright (C) 2014, Sugar Labs
# Copyright (C) 2014, Walter Bender
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from gi.repository import Gtk
from gi.repository import Gio
from gi.repository import GObject

from sugar3.graphics.icon import EventIcon
from sugar3.graphics import style
from sugar3.graphics.xocolor import XoColor


GENDERS = ['female', 'male']


def load_gender():
    settings = Gio.Settings.new('org.sugarlabs.user')
    return settings.get_string('gender')


def save_gender(gender):
    settings = Gio.Settings.new('org.sugarlabs.user')
    if gender is not None:
        settings.set_string('gender', gender)
    else:
        settings.set_string('gender', '')


class GenderPicker(Gtk.Grid):

    gender_changed_signal = GObject.Signal('gender-changed', arg_types=([str]))

    def __init__(self):
        Gtk.Grid.__init__(self)
        self.set_row_spacing(style.DEFAULT_SPACING)
        self.set_column_spacing(style.DEFAULT_SPACING)

        self._gender = load_gender()
        self._buttons = []
        self._nocolor = XoColor('#010101,#ffffff')
        self._color = XoColor()

        for i, gender in enumerate(GENDERS):
            self._buttons.append(EventIcon(pixel_size=style.XLARGE_ICON_SIZE,
                                           icon_name='%s-6' % (gender)))
            self._buttons[-1].connect('activate',
                                      self._button_activate_cb, i)
            self.attach(self._buttons[-1], i * 2, 0, 1, 1)
            self._buttons[-1].show()

        self.reset_button = EventIcon(pixel_size=style.SMALL_ICON_SIZE,
                                      icon_name='entry-cancel')
        self.reset_button.connect('activate',
                                  self._reset_button_activate_cb)
        self.attach(self.reset_button, 1, 0, 1, 1)
        self.reset_button.xo_color = XoColor('#010101,#a0a0a0')
        self.reset_button.show()

    def _reset_button_activate_cb(self, widget):
        self._set_gender('')
        for i in range(len(GENDERS)):
            self._buttons[i].xo_color = self._nocolor

    def _button_activate_cb(self, widget, gender_index):
        self._set_gender(GENDERS[gender_index])
        self._buttons[gender_index].xo_color = self._color
        self._buttons[1 - gender_index].xo_color = self._nocolor

    def get_gender(self):
        return self._gender

    def _set_gender(self, gender):
        self.gender_changed_signal.emit(gender)
        self._gender = gender

    def update_color(self, color):
        self._color = color
        if self._gender in GENDERS:
            self._buttons[GENDERS.index(self._gender)].xo_color = self._color



# Copyright (C) 2007, Red Hat, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk
from gi.repository import Gdk
from gi.repository import GObject
from gi.repository import GLib
from gi.repository import Wnck

from sugar3.graphics import style


_MAX_DELAY = 1000

_CORNERS = ['nw', 'ne', 'se', 'sw']
_EDGES = ['n', 'e', 's', 'w']
_BOXES = _CORNERS + _EDGES


class EventArea(GObject.GObject):
    __gsignals__ = {
        'enter': (GObject.SignalFlags.RUN_FIRST, None, ([])),
        'leave': (GObject.SignalFlags.RUN_FIRST, None, ([])),
    }

    def __init__(self, settings):
        GObject.GObject.__init__(self)

        self._hover = False
        self._sids = {}

        self._boxes = {}
        self._tags = {}
        for tag in _BOXES:
            box = self._box(tag)
            self._tags[box] = tag
            self._boxes[tag] = box

        settings.connect('changed', self._settings_changed_cb)
        self._settings_changed_cb(settings, None)

        screen = Wnck.Screen.get_default()
        screen.connect('window-stacking-changed',
                       self._window_stacking_changed_cb)

    def _box(self, tag):
        box = Gtk.Invisible()
        box.connect('enter-notify-event', self._enter_notify_cb)
        box.connect('leave-notify-event', self._leave_notify_cb)
        box.drag_dest_set(0, [], 0)
        box.connect('drag-motion', self._drag_motion_cb)
        box.connect('drag-leave', self._drag_leave_cb)
        box.realize()
        return box

    def _settings_changed_cb(self, settings, key):
        self._edge_delay = min(settings.get_int('edge-delay'), _MAX_DELAY)
        self._corner_delay = min(settings.get_int('corner-delay'), _MAX_DELAY)
        ts = min(settings.get_int('trigger-size'), style.GRID_CELL_SIZE)
        sw = Gdk.Screen.width()
        sh = Gdk.Screen.height()

        if self._edge_delay == _MAX_DELAY:
            self._hide(_EDGES)
        else:
            self._move('n', ts, -1, sw - ts * 2, ts + 1)
            self._move('e', -1, ts, ts + 1, sh - ts * 2)
            self._move('s', ts, sh - ts, sw - ts * 2, ts + 1)
            self._move('w', sw - ts, ts, ts + 1, sh - ts * 2)

        if self._corner_delay == _MAX_DELAY:
            self._hide(_CORNERS)
        else:
            self._move('nw', -1, -1, ts + 1, ts + 1)
            self._move('ne', sw - ts, -1, ts + 1, ts + 1)
            self._move('se', sw - ts, sh - ts, ts + 1, ts + 1)
            self._move('sw', -1, sh - ts, ts + 1, ts + 1)

    def _hide(self, tags):
        for tag in tags:
            self._move(tag, -20, -20, 1, 1)

    def _move(self, tag, x, y, width, height):
        window = self._boxes[tag].get_window()
        window.set_events(Gdk.EventMask.POINTER_MOTION_MASK |
                          Gdk.EventMask.ENTER_NOTIFY_MASK |
                          Gdk.EventMask.LEAVE_NOTIFY_MASK)
        window.move_resize(x, y, width, height)

    def _notify_enter(self):
        if not self._hover:
            self._hover = True
            self.emit('enter')

    def _notify_leave(self):
        if self._hover:
            self._hover = False
            self.emit('leave')

    def _enter_notify_cb(self, widget, event):
        if self._sids:
            GLib.source_remove(widget.sid)
            del self._sids[widget]

        delay = None
        if self._tags[widget] in _CORNERS:
            delay = self._corner_delay
        if self._tags[widget] in _EDGES:
            delay = self._edge_delay

        if delay is not None:
            self._sids[widget] = GLib.timeout_add(delay,
                                                  self.__delay_cb,
                                                  widget)

    def __delay_cb(self, widget):
        del self._sids[widget]
        self._notify_enter()
        return False

    def _leave_notify_cb(self, widget, event):
        if widget in self._sids:
            GLib.source_remove(self._sids[widget])
            del self._sids[widget]
        self._notify_leave()

    def _drag_motion_cb(self, widget, drag_context, x, y, timestamp):
        Gdk.drag_status(drag_context, 0, timestamp)
        self._notify_enter()
        return True

    def _drag_leave_cb(self, widget, drag_context, timestamp):
        self._notify_leave()
        return True

    def show(self):
        for box in list(self._boxes.values()):
            box.show()

    def hide(self):
        for box in list(self._boxes.values()):
            box.hide()

    def _window_stacking_changed_cb(self, screen):
        for box in list(self._boxes.values()):
            box.get_window().raise_()



# Copyright (C) 2006, Red Hat, Inc.
# Copyright (C) 2009, One Laptop Per Child Association Inc
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging

import gi
gi.require_version('Xkl', '1.0')
from gi.repository import Gio
from gi.repository import GdkX11
from gi.repository import Xkl


def setup():
    settings = Gio.Settings.new('org.sugarlabs.peripherals.keyboard')
    have_config = False

    try:
        display = GdkX11.x11_get_default_xdisplay()
        if display is not None:
            engine = Xkl.Engine.get_instance(display)
        else:
            logging.debug('setup_keyboard_cb: Could not get default display.')
            return

        configrec = Xkl.ConfigRec()
        configrec.get_from_server(engine)

        layouts = settings.get_strv('layouts')
        layouts_list = []
        variants_list = []
        if layouts:
            for layout in layouts:
                layouts_list.append(layout.split('(')[0])
                variants_list.append(layout.split('(')[1][:-1])

            if layouts_list and variants_list:
                have_config = True
                configrec.set_layouts(layouts_list)
                configrec.set_variants(variants_list)

        model = settings.get_string('model')
        if model:
            have_config = True
            configrec.set_model(model)

        options = settings.get_strv('options')
        if options:
            have_config = True
            configrec.set_options(options)

        if have_config:
            configrec.activate(engine)
    except Exception:
        logging.exception('Error during keyboard configuration')



# Copyright (C) 2010, Walter Bender, Sugar Labs
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from gettext import gettext as _
import os

from gi.repository import Gtk

import logging

from sugar3 import profile
from sugar3.graphics.tray import TrayIcon
from sugar3.graphics.palette import Palette
from sugar3.graphics import style

from jarabe.frame.frameinvoker import FrameWidgetInvoker

TOUCHPAD_MODE_MOUSE = 'mouse'
TOUCHPAD_MODE_PENTABLET = 'pentablet'

TOUCHPAD_MODES = (TOUCHPAD_MODE_MOUSE, TOUCHPAD_MODE_PENTABLET)
STATUS_TEXT = (_('finger'), _('stylus'))
STATUS_ICON = ('touchpad-capacitive', 'touchpad-resistive')

# NODE_PATH is used to communicate with the touchpad device.
NODE_PATH = '/sys/devices/platform/i8042/serio1/hgpk_mode'


class DeviceView(TrayIcon):
    """ Manage the touchpad mode from the device palette on the Frame. """

    FRAME_POSITION_RELATIVE = 500

    def __init__(self):
        """ Create the icon that represents the touchpad. """
        icon_name = STATUS_ICON[_read_touchpad_mode()]

        color = profile.get_color()
        TrayIcon.__init__(self, icon_name=icon_name, xo_color=color)

        self.set_palette_invoker(FrameWidgetInvoker(self))
        self.connect('button-release-event', self.__button_release_event_cb)

    def create_palette(self):
        """ Create a palette for this icon; called by the Sugar framework
        when a palette needs to be displayed. """
        self.palette = ResourcePalette(_('My touchpad'), self.icon)
        self.palette.set_group_id('frame')
        return self.palette

    def __button_release_event_cb(self, widget, event):
        """ Callback for button release event; used to invoke touchpad-mode
        change. """
        self.palette.toggle_mode()
        return True


class ResourcePalette(Palette):
    """ Palette attached to the decive icon that represents the touchpas. """

    def __init__(self, primary_text, icon):
        """ Create the palette and initilize with current touchpad status. """
        Palette.__init__(self, label=primary_text)

        self._icon = icon

        vbox = Gtk.VBox()
        self.set_content(vbox)

        self._status_text = Gtk.Label()
        vbox.pack_start(self._status_text, True, True, style.DEFAULT_PADDING)
        self._status_text.show()

        vbox.show()

        self._mode = _read_touchpad_mode()
        self._update()

    def _update(self):
        """ Update the label and icon based on the current mode. """
        self._status_text.set_label(STATUS_TEXT[self._mode])
        self._icon.props.icon_name = STATUS_ICON[self._mode]

    def toggle_mode(self):
        """ Toggle the touchpad mode. """
        self._mode = 1 - self._mode
        _write_touchpad_mode(self._mode)
        self._update()


def setup(tray):
    """ Initialize the devic icon; called by the shell when initializing the
    Frame. """
    if os.path.exists(NODE_PATH):
        tray.add_device(DeviceView())
        _write_touchpad_mode_str(TOUCHPAD_MODE_MOUSE)


def _read_touchpad_mode_str():
    """ Read the touchpad mode string from the node path. """
    node_file_handle = open(NODE_PATH, 'r')
    text = node_file_handle.read().strip().lower()
    node_file_handle.close()
    return text


def _read_touchpad_mode():
    """ Read the touchpad mode and return the mode index. """
    mode_str = _read_touchpad_mode_str()
    if mode_str not in TOUCHPAD_MODES:
        return None
    return TOUCHPAD_MODES.index(mode_str)


def _write_touchpad_mode_str(mode_str):
    """ Write the touchpad mode to the node path. """
    try:
        node_file_handle = open(NODE_PATH, 'w')
    except IOError as e:
        logging.error('Error opening %s for writing: %s', NODE_PATH, e)
        return
    node_file_handle.write(mode_str)
    node_file_handle.close()


def _write_touchpad_mode(mode_num):
    """ Look up the mode (by index) and write to node path. """
    return _write_touchpad_mode_str(TOUCHPAD_MODES[mode_num])



# Copyright (C) 2007, Red Hat, Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GObject

from sugar3.graphics import style
from sugar3.graphics import animator
from sugar3.graphics.icon import Icon

from jarabe.model.buddy import get_owner_instance
from jarabe.desktop.viewcontainer import ViewContainer
from jarabe.desktop.favoriteslayout import SpreadLayout


class _Animation(animator.Animation):

    def __init__(self, icon, start_size, end_size):
        animator.Animation.__init__(self, 0.0, 1.0)

        self._icon = icon
        self.start_size = start_size
        self.end_size = end_size

    def next_frame(self, current):
        d = (self.end_size - self.start_size) * current
        self._icon.props.pixel_size = int(self.start_size + d)


class TransitionBox(ViewContainer):
    __gtype_name__ = 'SugarTransitionBox'

    __gsignals__ = {
        'completed': (GObject.SignalFlags.RUN_FIRST, None, ([])),
    }

    def __init__(self):
        layout = SpreadLayout()

        # Round off icon size to an even number to ensure that the icon
        owner = get_owner_instance()
        self._owner_icon = Icon(icon_name='computer-xo',
                                xo_color=owner.get_color(),
                                pixel_size=style.XLARGE_ICON_SIZE & ~1)
        ViewContainer.__init__(self, layout, self._owner_icon)

        self._animator = animator.Animator(0.3, widget=self)
        self._animator.connect('completed', self._animation_completed_cb)

    def _animation_completed_cb(self, anim):
        self.emit('completed')

    def start_transition(self, start_size, end_size):
        self._animator.remove_all()
        self._animator.add(_Animation(self._owner_icon, start_size, end_size))
        self._animator.start()



# Copyright (C) 2008 One Laptop Per Child
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import textwrap
from gettext import gettext as _

from gi.repository import GObject
from gi.repository import GLib
from gi.repository import Gtk
from gi.repository import Gdk

from sugar3 import profile
from sugar3.graphics import style
from sugar3.graphics.xocolor import XoColor
from sugar3.graphics.icon import Icon
from sugar3.graphics.icon import get_surface
from sugar3.graphics.palette import Palette
from sugar3.graphics.palettemenu import PaletteMenuItem
from sugar3.graphics.palettemenu import PaletteMenuItemSeparator
from sugar3.graphics.toolbutton import ToolButton

from jarabe.model import notifications
from jarabe.view.pulsingicon import PulsingIcon
from jarabe.frame.frameinvoker import FrameWidgetInvoker


class NotificationBox(Gtk.VBox):

    LINES = 3
    MAX_ENTRIES = 3
    ELLIPSIS_AND_BREAKS = 6

    def __init__(self, name):
        Gtk.VBox.__init__(self)
        self._name = name

        self._notifications_box = Gtk.VBox()
        self._notifications_box.show()

        self._scrolled_window = Gtk.ScrolledWindow()
        self._scrolled_window.add_with_viewport(self._notifications_box)
        self._scrolled_window.set_policy(Gtk.PolicyType.NEVER,
                                         Gtk.PolicyType.AUTOMATIC)
        self._scrolled_window.show()

        separator = PaletteMenuItemSeparator()
        separator.show()

        clear_item = PaletteMenuItem(_('Clear notifications'), 'dialog-cancel')
        clear_item.connect('activate', self.__clear_cb)
        clear_item.show()

        self.add(self._scrolled_window)
        self.add(separator)
        self.add(clear_item)

        self._service = notifications.get_service()
        entries = self._service.retrieve_by_name(self._name)

        if entries:
            for entry in entries:
                self._add(entry['summary'], entry['body'])

        self._service.notification_received.connect(
            self.__notification_received_cb)

        self.connect('destroy', self.__destroy_cb)

    def _update_scrolled_size(self):
        entries = self._notifications_box.get_children()

        height = 0
        for entry in entries[:self.MAX_ENTRIES]:
            requests = entry.get_preferred_size()
            height += requests[1].height

        self._scrolled_window.set_size_request(-1, height)

    def _add(self, summary, body):
        icon = Icon()
        icon.props.icon_name = 'emblem-notification'
        icon.props.icon_size = Gtk.IconSize.SMALL_TOOLBAR
        icon.props.xo_color = \
            XoColor('%s,%s' % (style.COLOR_WHITE.get_svg(),
                               style.COLOR_BLACK.get_svg()))
        icon.show()

        summary_label = Gtk.Label()
        summary_label.set_max_width_chars(style.MENU_WIDTH_CHARS)
        summary_label.set_ellipsize(style.ELLIPSIZE_MODE_DEFAULT)
        summary_label.set_alignment(0, 0.5)
        summary_label.set_markup('<b>%s</b>' % summary)
        summary_label.show()

        body_label = Gtk.Label()
        body_label.set_alignment(0, 0.5)

        if hasattr(body_label, 'set_lines'):
            body_label.set_max_width_chars(style.MENU_WIDTH_CHARS)
            body_label.set_line_wrap(True)
            body_label.set_ellipsize(style.ELLIPSIZE_MODE_DEFAULT)
            body_label.set_lines(self.LINES)
            body_label.set_justify(Gtk.Justification.FILL)
        else:
            # FIXME: fallback for Gtk < 3.10
            body_width = self.LINES * style.MENU_WIDTH_CHARS
            body_width -= self.ELLIPSIS_AND_BREAKS
            body = body.replace('\n', ' ')
            if len(body) > body_width:
                body = ' '.join(body[:body_width].split(' ')[:-1]) + '...'
            body = textwrap.fill(body, width=style.MENU_WIDTH_CHARS)

        body_label.set_text(body)
        body_label.show()

        grid = Gtk.Grid()
        grid.set_border_width(style.DEFAULT_SPACING)
        grid.set_column_spacing(style.DEFAULT_SPACING)
        grid.set_row_spacing(0)
        grid.attach(icon, 0, 0, 1, 2)
        grid.attach(summary_label, 1, 0, 1, 1)
        grid.attach(body_label, 1, 1, 1, 1)
        grid.show()

        self._notifications_box.add(grid)
        self._update_scrolled_size()
        self.show()

    def __clear_cb(self, clear_item):
        logging.debug('NotificationBox.__clear_cb')
        for entry in self._notifications_box.get_children():
            self._notifications_box.remove(entry)
        self._service.clear_by_name(self._name)
        self.hide()

    def __notification_received_cb(self, **kwargs):
        logging.debug('NotificationBox.__notification_received_cb')
        if kwargs.get('app_name', '') == self._name:
            self._add(kwargs.get('summary', ''), kwargs.get('body', ''))

    def __destroy_cb(self, box):
        logging.debug('NotificationBox.__destroy_cb')
        service = notifications.get_service()
        service.notification_received.disconnect(
            self. __notification_received_cb)


class NotificationButton(ToolButton):

    def __init__(self, name):
        ToolButton.__init__(self)
        self._name = name
        self._icon = None
        self.set_palette_invoker(FrameWidgetInvoker(self))
        self.palette_invoker.cache_palette = False
        self.connect('clicked', self.__clicked_cb)

    def set_icon(self, icon):
        self._icon = icon
        self._icon.show()
        self.set_icon_widget(self._icon)

    def show_badge(self):
        if self._icon:
            self._icon.show_badge()

    def hide_badge(self):
        if self._icon:
            self._icon.hide_badge()

    def create_palette(self):
        notification_box = NotificationBox(self._name)
        palette = Palette(self._name)
        palette.set_group_id('frame')
        palette.set_content(notification_box)
        self.set_palette(palette)

    def __clicked_cb(self, button):
        self.create_palette()
        self.palette.popup(immediate=True)


class NotificationPulsingIcon(PulsingIcon):

    def __init__(self, filename=None, name=None, colors=None):
        PulsingIcon.__init__(self, pixel_size=style.STANDARD_ICON_SIZE)
        self._badge = None

        if filename:
            self.props.file = filename
        elif name:
            self.props.icon_name = name
        else:
            self.props.icon_name = 'application-octet-stream'

        if not colors:
            colors = profile.get_color()
        self.props.base_color = colors
        self.props.pulse_color = \
            XoColor('%s,%s' % (style.COLOR_BUTTON_GREY.get_svg(),
                               style.COLOR_TOOLBAR_GREY.get_svg()))

    def show_badge(self):
        self._badge = get_surface(icon_name='emblem-notification',
                                  stroke_color=style.COLOR_WHITE.get_svg(),
                                  fill_color=style.COLOR_BLACK.get_svg(),
                                  width=self.get_badge_size(),
                                  height=self.get_badge_size())

    def hide_badge(self):
        self._badge = None

    def do_draw(self, cr):
        PulsingIcon.do_draw(self, cr)
        if self._badge:
            allocation = self.get_allocation()

            # XXX assume icon is centered in its container
            offset = int(self.props.pixel_size / 2) - self.get_badge_size()
            x = int(allocation.width / 2) + offset
            y = int(allocation.height / 2) + offset

            cr.set_source_surface(self._badge, x, y)
            cr.paint()


class NotificationIcon(Gtk.EventBox):
    __gtype_name__ = 'SugarNotificationIcon'

    __gproperties__ = {
        'xo-color': (object, None, None, GObject.ParamFlags.READWRITE),
        'icon-name': (str, None, None, None, GObject.ParamFlags.READWRITE),
        'icon-filename': (str, None, None, None, GObject.ParamFlags.READWRITE),
    }

    _PULSE_TIMEOUT = 3

    def __init__(self, **kwargs):
        self._icon = NotificationPulsingIcon()
        self._icon.props.pixel_size = style.STANDARD_ICON_SIZE

        Gtk.EventBox.__init__(self, **kwargs)
        self.props.visible_window = False

        self._icon.props.pulse_color = \
            XoColor('%s,%s' % (style.COLOR_BUTTON_GREY.get_svg(),
                               style.COLOR_TRANSPARENT.get_svg()))
        self._icon.props.pulsing = True
        self.add(self._icon)
        self._icon.show()

        GLib.timeout_add_seconds(self._PULSE_TIMEOUT,
                                 self.__stop_pulsing_cb)

        self.set_size_request(style.GRID_CELL_SIZE, style.GRID_CELL_SIZE)

    def __stop_pulsing_cb(self):
        self._icon.props.pulsing = False
        return False

    def do_set_property(self, pspec, value):
        if pspec.name == 'xo-color':
            if self._icon.props.base_color != value:
                self._icon.props.base_color = value
        elif pspec.name == 'icon-name':
            if self._icon.props.icon_name != value:
                self._icon.props.icon_name = value
        elif pspec.name == 'icon-filename':
            if self._icon.props.file != value:
                self._icon.props.file = value

    def do_get_property(self, pspec):
        if pspec.name == 'xo-color':
            return self._icon.props.base_color
        if pspec.name == 'icon-name':
            return self._icon.props.icon_name
        if pspec.name == 'icon-filename':
            return self._icon.props.file

    def _set_palette(self, palette):
        self._icon.palette = palette

    def _get_palette(self):
        return self._icon.palette

    palette = property(_get_palette, _set_palette)

    def show_badge(self):
        self._icon.show_badge()

    def hide_badge(self):
        self._icon.hide_badge()


class NotificationWindow(Gtk.Window):
    __gtype_name__ = 'SugarNotificationWindow'

    def __init__(self, **kwargs):

        Gtk.Window.__init__(self, **kwargs)

        self.set_decorated(False)
        self.set_resizable(False)
        self.connect('realize', self._realize_cb)

    def _realize_cb(self, widget):
        self.set_type_hint(Gdk.WindowTypeHint.DIALOG)
        self.get_window().set_accept_focus(False)

        color = Gdk.color_parse(style.COLOR_TOOLBAR_GREY.get_html())
        self.modify_bg(Gtk.StateType.NORMAL, color)



# Copyright (C) 2008 One Laptop Per Child
# Copyright (C) 2014 Ignacio Rodriguez
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gettext import gettext as _
from gettext import ngettext
import logging
import os

from gi.repository import GObject
from gi.repository import GLib
from gi.repository import Gtk
from gi.repository import Gdk
from gi.repository import Gio
from gi.repository import SugarExt

from sugar3.graphics import style
from sugar3.graphics.palette import Palette
from sugar3.graphics.menuitem import MenuItem
from sugar3.graphics.icon import Icon
from sugar3.graphics.xocolor import XoColor
from sugar3.graphics.alert import Alert
from sugar3 import mime
from sugar3 import profile

from jarabe.model import friends
from jarabe.model import filetransfer
from jarabe.model import mimeregistry
from jarabe.journal import misc
from jarabe.journal import model
from jarabe.journal import journalwindow
from jarabe.webservice import accountsmanager
from jarabe.journal.misc import get_mount_color

PROJECT_BUNDLE_ID = 'org.sugarlabs.Project'


class ObjectPalette(Palette):

    __gtype_name__ = 'ObjectPalette'

    __gsignals__ = {
        'detail-clicked': (GObject.SignalFlags.RUN_FIRST, None,
                           ([str])),
        'volume-error': (GObject.SignalFlags.RUN_FIRST, None,
                         ([str, str])),
        'choose-project': (GObject.SignalFlags.RUN_FIRST, None,
                           ([object])),
    }

    def __init__(self, journalactivity, metadata, detail=False):

        self._journalactivity = journalactivity
        self._metadata = metadata

        activity_icon = Icon(pixel_size=style.STANDARD_ICON_SIZE)
        activity_icon.props.file = misc.get_icon_name(metadata)
        color = misc.get_icon_color(metadata)
        activity_icon.props.xo_color = color

        if 'title' in metadata:
            title = metadata['title']
        else:
            title = _('Untitled')

        Palette.__init__(self, primary_text=title,
                         icon=activity_icon)

        description = metadata.get('description', '')
        if description:
            self.set_secondary_text(description)

        if misc.can_resume(metadata):
            if metadata.get('activity_id', ''):
                resume_label = _('Resume')
                resume_with_label = _('Resume with')
            else:
                resume_label = _('Start')
                resume_with_label = _('Start with')
            menu_item = MenuItem(resume_label, 'activity-start')
            menu_item.connect('activate', self.__start_activate_cb)
            self.menu.append(menu_item)
            menu_item.show()

            menu_item = MenuItem(resume_with_label, 'activity-start')
            self.menu.append(menu_item)
            menu_item.show()
            start_with_menu = StartWithMenu(self._metadata)
            menu_item.set_submenu(start_with_menu)

        elif metadata.get('activity', None) == PROJECT_BUNDLE_ID:
            open_label = _('Open')
            menu_item = MenuItem(open_label, 'project-box')
            menu_item.connect('activate', self.__open_project_activate_cb)
            self.menu.append(menu_item)
            menu_item.show()

        else:
            menu_item = MenuItem(_('No activity to start entry'))
            menu_item.set_sensitive(False)
            self.menu.append(menu_item)
            menu_item.show()

        menu_item = MenuItem(_('Copy to'))
        icon = Icon(icon_name='edit-copy', xo_color=color,
                    pixel_size=style.SMALL_ICON_SIZE)
        menu_item.set_image(icon)
        self.menu.append(menu_item)
        menu_item.show()
        copy_menu = CopyMenu(self._journalactivity, self.__get_uid_list_cb)
        copy_menu.connect('volume-error', self.__volume_error_cb)
        menu_item.set_submenu(copy_menu)

        if not metadata.get('activity', None) == PROJECT_BUNDLE_ID:
            menu_item = MenuItem(_('Send to project...'), 'project-box')
            menu_item.connect('activate', self.__copy_to_project_activated_cb)
            self.menu.append(menu_item)
            menu_item.show()

        if self._metadata['mountpoint'] == '/':
            menu_item = MenuItem(_('Duplicate'))
            icon = Icon(icon_name='edit-duplicate', xo_color=color,
                        pixel_size=style.SMALL_ICON_SIZE)
            menu_item.set_image(icon)
            menu_item.connect('activate', self.__duplicate_activate_cb)
            self.menu.append(menu_item)
            menu_item.show()

        menu_item = MenuItem(_('Send to'), 'document-send')
        self.menu.append(menu_item)
        menu_item.show()

        friends_menu = FriendsMenu()
        friends_menu.connect('friend-selected', self.__friend_selected_cb)
        menu_item.set_submenu(friends_menu)

        if detail is True:
            menu_item = MenuItem(_('View Details'), 'go-right')
            menu_item.connect('activate', self.__detail_activate_cb)
            self.menu.append(menu_item)
            menu_item.show()

        menu_item = MenuItem(_('Erase'), 'list-remove')
        menu_item.connect('activate', self.__erase_activate_cb)
        self.menu.append(menu_item)
        menu_item.show()

    def __get_uid_list_cb(self):
        return [self._metadata['uid']]

    def __copy_to_project_activated_cb(self, menu_item):
        self.emit('choose-project', self._metadata)
        self.destroy()

    def __open_project_activate_cb(self, menu_item):
        self._journalactivity.project_view_activated_cb(
            list_view=None,
            metadata=self._metadata)

    def __start_activate_cb(self, menu_item):
        misc.resume(self._metadata,
                    alert_window=journalwindow.get_journal_window())

    def __duplicate_activate_cb(self, menu_item):
        try:
            model.copy(self._metadata, '/')
        except IOError as e:
            logging.exception('Error while copying the entry. %s', e.strerror)
            self.emit('volume-error',
                      _('Error while copying the entry. %s') % e.strerror,
                      _('Error'))

    def __erase_activate_cb(self, menu_item):
        alert = Alert()
        erase_string = _('Erase')
        alert.props.title = erase_string
        alert.props.msg = _('Do you want to permanently erase \"%s\"?') \
            % self._metadata['title']
        icon = Icon(icon_name='dialog-cancel')
        alert.add_button(Gtk.ResponseType.CANCEL, _('Cancel'), icon)
        icon.show()
        ok_icon = Icon(icon_name='dialog-ok')
        alert.add_button(Gtk.ResponseType.OK, erase_string, ok_icon)
        ok_icon.show()
        alert.connect('response', self.__erase_alert_response_cb)
        journalwindow.get_journal_window().add_alert(alert)
        alert.show()

    def __erase_alert_response_cb(self, alert, response_id):
        journalwindow.get_journal_window().remove_alert(alert)
        if response_id is Gtk.ResponseType.OK:
            uid = self._metadata['uid']
            list_model = self._journalactivity.get_list_view().get_model()
            if list_model.is_selected(uid):
                list_model.set_selected(uid, False)
            model.delete(uid)

    def __detail_activate_cb(self, menu_item):
        self.emit('detail-clicked', self._metadata['uid'])

    def __volume_error_cb(self, menu_item, message, severity):
        self.emit('volume-error', message, severity)

    def __friend_selected_cb(self, menu_item, buddy):
        logging.debug('__friend_selected_cb')
        file_name = model.get_file(self._metadata['uid'])

        if not file_name or not os.path.exists(file_name):
            logging.warn('Entries without a file cannot be sent.')
            self.emit('volume-error',
                      _('Entries without a file cannot be sent.'),
                      _('Warning'))
            return

        title = str(self._metadata['title'])
        description = str(self._metadata.get('description', ''))
        mime_type = str(self._metadata['mime_type'])

        if not mime_type:
            mime_type = mime.get_for_file(file_name)

        filetransfer.start_transfer(buddy, file_name, title, description,
                                    mime_type)

    def popup(self, immediate=False, state=None):
        if self._journalactivity.get_list_view().is_dragging():
            return

        Palette.popup(self, immediate)


class CopyMenu(Gtk.Menu):
    __gtype_name__ = 'JournalCopyMenu'

    __gsignals__ = {
        'volume-error': (GObject.SignalFlags.RUN_FIRST, None,
                         ([str, str])),
    }

    def __init__(self, journalactivity, get_uid_list_cb):
        Gtk.Menu.__init__(self)
        CopyMenuBuilder(journalactivity, get_uid_list_cb,
                        self.__volume_error_cb, self)

    def __volume_error_cb(self, menu_item, message, severity):
        self.emit('volume-error', message, severity)


class CopyMenuBuilder():

    def __init__(self, journalactivity, get_uid_list_cb, __volume_error_cb,
                 menu, add_clipboard_menu=True, add_webservices_menu=True):

        self._journalactivity = journalactivity
        self._get_uid_list_cb = get_uid_list_cb
        self.__volume_error_cb = __volume_error_cb
        self._menu = menu
        self._add_clipboard_menu = add_clipboard_menu
        self._add_webservices_menu = add_webservices_menu

        self._mount_added_hid = None
        self._mount_removed_hid = None
        self._create_menu_items()

    def _create_menu_items(self):
        if self._add_clipboard_menu:
            clipboard_menu = ClipboardMenu(self._get_uid_list_cb)
            clipboard_menu.set_image(Icon(icon_name='toolbar-edit',
                                          pixel_size=style.SMALL_ICON_SIZE))
            clipboard_menu.connect('volume-error', self.__volume_error_cb)
            self._menu.append(clipboard_menu)
            clipboard_menu.show()

        if self._journalactivity.get_mount_point() != '/':
            color = profile.get_color()
            journal_menu = VolumeMenu(self._journalactivity,
                                      self._get_uid_list_cb, _('Journal'), '/')
            journal_menu.set_image(Icon(icon_name='activity-journal',
                                        xo_color=color,
                                        pixel_size=style.SMALL_ICON_SIZE))
            journal_menu.connect('volume-error', self.__volume_error_cb)
            self._menu.append(journal_menu)
            journal_menu.show()

        documents_path = model.get_documents_path()
        if documents_path is not None and \
                self._journalactivity.get_mount_point() != documents_path:
            documents_menu = VolumeMenu(self._journalactivity,
                                        self._get_uid_list_cb, _('Documents'),
                                        documents_path)
            documents_menu.set_image(Icon(icon_name='user-documents',
                                          pixel_size=style.SMALL_ICON_SIZE))
            documents_menu.connect('volume-error', self.__volume_error_cb)
            self._menu.append(documents_menu)
            documents_menu.show()

        volume_monitor = Gio.VolumeMonitor.get()
        self._volumes = {}
        for mount in volume_monitor.get_mounts():
            self._add_mount(mount)

        self._mount_added_hid = volume_monitor.connect('mount-added',
                                                       self.__mount_added_cb)
        self._mount_removed_hid = volume_monitor.connect(
            'mount-removed',
            self.__mount_removed_cb)

        if self._add_webservices_menu:
            for account in accountsmanager.get_configured_accounts():
                if hasattr(account, 'get_shared_journal_entry'):
                    entry = account.get_shared_journal_entry()
                    if hasattr(entry, 'get_share_menu'):
                        self._menu.append(entry.get_share_menu(
                                          self._get_uid_list_cb))

    def update_mount_point(self):
        for menu_item in self._menu.get_children():
            if isinstance(menu_item, MenuItem):
                self._menu.remove(menu_item)
        self._create_menu_items()

    def __mount_added_cb(self, volume_monitor, mount):
        self._add_mount(mount)

    def _add_mount(self, mount):
        mount_path = mount.get_root().get_path()
        if mount_path in self._volumes:
            return
        if self._journalactivity.get_mount_point() == mount_path:
            return
        volume_menu = VolumeMenu(self._journalactivity,
                                 self._get_uid_list_cb, mount.get_name(),
                                 mount.get_root().get_path())
        icon_name = misc.get_mount_icon_name(mount, Gtk.IconSize.MENU)
        icon = Icon(pixel_size=style.SMALL_ICON_SIZE,
                    icon_name=icon_name,
                    xo_color=get_mount_color(mount))

        volume_menu.set_image(icon)
        volume_menu.connect('volume-error', self.__volume_error_cb)
        self._menu.append(volume_menu)
        self._volumes[mount.get_root().get_path()] = volume_menu
        volume_menu.show()

    def __mount_removed_cb(self, volume_monitor, mount):
        volume_menu = self._volumes[mount.get_root().get_path()]
        self._menu.remove(volume_menu)
        del self._volumes[mount.get_root().get_path()]


class VolumeMenu(MenuItem):
    __gtype_name__ = 'JournalVolumeMenu'

    __gsignals__ = {
        'volume-error': (GObject.SignalFlags.RUN_FIRST, None,
                         ([str, str])),
    }

    def __init__(self, journalactivity, get_uid_list_cb, label, mount_point):
        MenuItem.__init__(self, label)
        self._get_uid_list_cb = get_uid_list_cb
        self._journalactivity = journalactivity
        self._mount_point = mount_point
        self.connect('activate', self.__copy_to_volume_cb)

    def __copy_to_volume_cb(self, menu_item):
        uid_list = self._get_uid_list_cb()
        if len(uid_list) == 1:
            uid = uid_list[0]
            file_path = model.get_file(uid)

            if not file_path or not os.path.exists(file_path):
                logging.warn('Entries without a file cannot be copied.')
                self.emit('volume-error',
                          _('Entries without a file cannot be copied.'),
                          _('Warning'))
                return

            try:
                metadata = model.get(uid)
                model.copy(metadata, self._mount_point)
            except IOError as e:
                logging.exception('Error while copying the entry. %s',
                                  e.strerror)
                self.emit('volume-error',
                          _('Error while copying the entry. %s') % e.strerror,
                          _('Error'))
        else:
            BatchOperator(
                self._journalactivity, uid_list, _('Copy'),
                self._get_confirmation_alert_message(len(uid_list)),
                self._perform_copy)

    def _get_confirmation_alert_message(self, entries_len):
        return ngettext('Do you want to copy %d entry?',
                        'Do you want to copy %d entries?',
                        entries_len) % (entries_len)

    def _perform_copy(self, metadata):
        file_path = model.get_file(metadata['uid'])
        if not file_path or not os.path.exists(file_path):
            logging.warn('Entries without a file cannot be copied.')
            return
        try:
            model.copy(metadata, self._mount_point)
        except IOError as e:
            logging.exception('Error while copying the entry. %s',
                              e.strerror)


class ClipboardMenu(MenuItem):
    __gtype_name__ = 'JournalClipboardMenu'

    __gsignals__ = {
        'volume-error': (GObject.SignalFlags.RUN_FIRST, None,
                         ([str, str])),
    }

    def __init__(self, get_uid_list_cb):
        MenuItem.__init__(self, _('Clipboard'))

        self._temp_file_path = None
        self._get_uid_list_cb = get_uid_list_cb
        self.connect('activate', self.__copy_to_clipboard_cb)

    def __copy_to_clipboard_cb(self, menu_item):
        clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
        uid_list = self._get_uid_list_cb()
        if len(uid_list) == 1:
            uid = uid_list[0]
            file_path = model.get_file(uid)
            if not file_path or not os.path.exists(file_path):
                logging.warn('Entries without a file cannot be copied.')
                self.emit('volume-error',
                          _('Entries without a file cannot be copied.'),
                          _('Warning'))
                return

            # XXX SL#4307 - until set_with_data bindings are fixed upstream
            if hasattr(clipboard, 'set_with_data'):
                clipboard.set_with_data(
                    [Gtk.TargetEntry.new('text/uri-list', 0, 0)],
                    self.__clipboard_get_func_cb,
                    self.__clipboard_clear_func_cb,
                    None)
            else:
                SugarExt.clipboard_set_with_data(
                    clipboard,
                    [Gtk.TargetEntry.new('text/uri-list', 0, 0)],
                    self.__clipboard_get_func_cb,
                    self.__clipboard_clear_func_cb,
                    None)

    def __clipboard_get_func_cb(self, clipboard, selection_data, info, data):
        # Get hold of a reference so the temp file doesn't get deleted
        for uid in self._get_uid_list_cb():
            self._temp_file_path = model.get_file(uid)
            logging.debug('__clipboard_get_func_cb %r', self._temp_file_path)
            selection_data.set_uris(['file://' + self._temp_file_path])

    def __clipboard_clear_func_cb(self, clipboard, data):
        # Release and delete the temp file
        self._temp_file_path = None


class FriendsMenu(Gtk.Menu):
    __gtype_name__ = 'JournalFriendsMenu'

    __gsignals__ = {
        'friend-selected': (GObject.SignalFlags.RUN_FIRST, None,
                            ([object])),
    }

    def __init__(self):
        Gtk.Menu.__init__(self)

        if filetransfer.file_transfer_available():
            friends_model = friends.get_model()
            for friend in friends_model:
                if friend.is_present():
                    menu_item = MenuItem(text_label=friend.get_nick(),
                                         icon_name='computer-xo',
                                         xo_color=friend.get_color())
                    menu_item.connect('activate', self.__item_activate_cb,
                                      friend)
                    self.append(menu_item)
                    menu_item.show()

            if not self.get_children():
                menu_item = MenuItem(_('No friends present'))
                menu_item.set_sensitive(False)
                self.append(menu_item)
                menu_item.show()
        else:
            menu_item = MenuItem(_('No valid connection found'))
            menu_item.set_sensitive(False)
            self.append(menu_item)
            menu_item.show()

    def __item_activate_cb(self, menu_item, friend):
        self.emit('friend-selected', friend)


class StartWithMenu(Gtk.Menu):
    __gtype_name__ = 'JournalStartWithMenu'

    def __init__(self, metadata):
        Gtk.Menu.__init__(self)

        self._metadata = metadata

        for activity_info in misc.get_activities(metadata):
            menu_item = MenuItem(activity_info.get_name())
            menu_item.set_image(Icon(file=activity_info.get_icon(),
                                     pixel_size=style.SMALL_ICON_SIZE))
            menu_item.connect('activate', self.__item_activate_cb,
                              activity_info.get_bundle_id())
            self.append(menu_item)
            menu_item.show()

        if not self.get_children():
            if metadata.get('activity_id', ''):
                resume_label = _('No activity to resume entry')
            else:
                resume_label = _('No activity to start entry')
            menu_item = MenuItem(resume_label)
            menu_item.set_sensitive(False)
            self.append(menu_item)
            menu_item.show()

    def __item_activate_cb(self, menu_item, service_name):
        mime_type = self._metadata.get('mime_type', '')
        if mime_type:
            mime_registry = mimeregistry.get_registry()
            mime_registry.set_default_activity(mime_type, service_name)
        misc.resume(self._metadata, bundle_id=service_name,
                    alert_window=journalwindow.get_journal_window())


class BuddyPalette(Palette):

    def __init__(self, buddy):
        self._buddy = buddy

        nick, colors = buddy
        buddy_icon = Icon(icon_name='computer-xo',
                          pixel_size=style.STANDARD_ICON_SIZE,
                          xo_color=XoColor(colors))

        Palette.__init__(self, primary_text=nick, icon=buddy_icon)

        # TODO: Support actions on buddies, like make friend, invite, etc.


class BatchOperator(GObject.GObject):
    """
    This class implements the course of actions that happens when clicking
    upon an BatchOperation  (eg. Batch-Copy-Toolbar-button;
                             Batch-Copy-To-Journal-button;
                             Batch-Copy-To-Documents-button;
                             Batch-Copy-To-Mounted-Drive-button;
                             Batch-Copy-To-Clipboard-button;
                             Batch-Erase-Button;
    """

    def __init__(self, journalactivity,
                 uid_list,
                 alert_title, alert_message,
                 operation_cb):
        GObject.GObject.__init__(self)

        self._journalactivity = journalactivity

        self._uid_list = uid_list[:]
        self._alert_title = alert_title
        self._alert_message = alert_message
        self._operation_cb = operation_cb

        self._show_confirmation_alert()

    def _show_confirmation_alert(self):
        self._journalactivity.freeze_ui()
        GLib.idle_add(self.__show_confirmation_alert_internal)

    def __show_confirmation_alert_internal(self):
        # Show a alert requesting confirmation before run the batch operation
        self._confirmation_alert = Alert()
        self._confirmation_alert.props.title = self._alert_title
        self._confirmation_alert.props.msg = self._alert_message

        stop_icon = Icon(icon_name='dialog-cancel')
        self._confirmation_alert.add_button(Gtk.ResponseType.CANCEL,
                                            _('Stop'), stop_icon)
        stop_icon.show()

        ok_icon = Icon(icon_name='dialog-ok')
        self._confirmation_alert.add_button(Gtk.ResponseType.OK,
                                            _('Continue'), ok_icon)
        ok_icon.show()

        self._journalactivity.add_alert(self._confirmation_alert)
        self._confirmation_alert.connect('response',
                                         self.__confirmation_response_cb)
        self._confirmation_alert.show()

    def __confirmation_response_cb(self, alert, response):
        if response == Gtk.ResponseType.CANCEL:
            self._journalactivity.unfreeze_ui()
            self._journalactivity.remove_alert(alert)
            # this is only in the case the operation already started
            # and the user want stop it.
            self._stop_batch_execution()
        elif not hasattr(self, '_object_index'):
            self._object_index = 0
            GLib.idle_add(self._operate_by_uid_internal)

    def _operate_by_uid_internal(self):
        # If there is still some uid left, proceed with the operation.
        # Else, proceed to post-operations.
        if self._object_index < len(self._uid_list):
            uid = self._uid_list[self._object_index]
            metadata = model.get(uid)
            title = None
            if 'title' in metadata:
                title = metadata['title']
            if title is None or title == '':
                title = _('Untitled')
            alert_message = _('%(index)d of %(total)d : %(object_title)s') % {
                'index': self._object_index + 1,
                'total': len(self._uid_list),
                'object_title': title}

            self._confirmation_alert.props.msg = alert_message
            GLib.idle_add(self._operate_per_metadata, metadata)
        else:
            self._finish_batch_execution()

    def _operate_per_metadata(self, metadata):
        self._operation_cb(metadata)

        # process the next
        self._object_index = self._object_index + 1
        GLib.idle_add(self._operate_by_uid_internal)

    def _stop_batch_execution(self):
        self._object_index = len(self._uid_list)

    def _finish_batch_execution(self):
        del self._object_index
        self._journalactivity.unfreeze_ui()
        self._journalactivity.remove_alert(self._confirmation_alert)
        self._journalactivity.update_selected_items_ui()



# Copyright (C) 2008 Red Hat, Inc.
# Copyright (C) 2009 Tomeu Vizoso, Simon Schampijer
# Copyright (C) 2009-2010 One Laptop per Child
# Copyright (C) 2009 Paraguay Educa, Martin Abente
# Copyright (C) 2010 Plan Ceibal, Daniel Castelo
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gettext import gettext as _
import logging
import os
import uuid

import dbus
import dbus.service
from gi.repository import GObject
import configparser
from gi.repository import Gio
import ctypes

from sugar3 import dispatch
from sugar3 import env

from jarabe.model.update.updater import check_urgent_update

NM_STATE_UNKNOWN = 0
NM_STATE_ASLEEP = 10
NM_STATE_DISCONNECTED = 20
NM_STATE_DISCONNECTING = 30
NM_STATE_CONNECTING = 40
NM_STATE_CONNECTED_LOCAL = 50
NM_STATE_CONNECTED_SITE = 60
NM_STATE_CONNECTED_GLOBAL = 70

NM_DEVICE_TYPE_UNKNOWN = 0
NM_DEVICE_TYPE_ETHERNET = 1
NM_DEVICE_TYPE_WIFI = 2
NM_DEVICE_TYPE_UNUSED1 = 3
NM_DEVICE_TYPE_UNUSED2 = 4
NM_DEVICE_TYPE_BT = 5
NM_DEVICE_TYPE_OLPC_MESH = 6
NM_DEVICE_TYPE_WIMAX = 7
NM_DEVICE_TYPE_MODEM = 8

NM_DEVICE_STATE_UNKNOWN = 0
NM_DEVICE_STATE_UNMANAGED = 10
NM_DEVICE_STATE_UNAVAILABLE = 20
NM_DEVICE_STATE_DISCONNECTED = 30
NM_DEVICE_STATE_PREPARE = 40
NM_DEVICE_STATE_CONFIG = 50
NM_DEVICE_STATE_NEED_AUTH = 60
NM_DEVICE_STATE_IP_CONFIG = 70
NM_DEVICE_STATE_IP_CHECK = 80
NM_DEVICE_STATE_SECONDARIES = 90
NM_DEVICE_STATE_ACTIVATED = 100
NM_DEVICE_STATE_DEACTIVATING = 110
NM_DEVICE_STATE_FAILED = 120

NM_CONNECTION_TYPE_802_11_OLPC_MESH = '802-11-olpc-mesh'
NM_CONNECTION_TYPE_802_11_WIRELESS = '802-11-wireless'
NM_CONNECTION_TYPE_GSM = 'gsm'

NM_ACTIVE_CONNECTION_STATE_UNKNOWN = 0
NM_ACTIVE_CONNECTION_STATE_ACTIVATING = 1
NM_ACTIVE_CONNECTION_STATE_ACTIVATED = 2
NM_ACTIVE_CONNECTION_STATE_DEACTIVATING = 3

NM_DEVICE_STATE_REASON_UNKNOWN = 0
NM_DEVICE_STATE_REASON_NONE = 1
NM_DEVICE_STATE_REASON_NOW_MANAGED = 2
NM_DEVICE_STATE_REASON_NOW_UNMANAGED = 3
NM_DEVICE_STATE_REASON_CONFIG_FAILED = 4
NM_DEVICE_STATE_REASON_IP_CONFIG_UNAVAILABLE = 5
NM_DEVICE_STATE_REASON_IP_CONFIG_EXPIRED = 6
NM_DEVICE_STATE_REASON_NO_SECRETS = 7
NM_DEVICE_STATE_REASON_SUPPLICANT_DISCONNECT = 8
NM_DEVICE_STATE_REASON_SUPPLICANT_CONFIG_FAILED = 9
NM_DEVICE_STATE_REASON_SUPPLICANT_FAILED = 10
NM_DEVICE_STATE_REASON_SUPPLICANT_TIMEOUT = 11
NM_DEVICE_STATE_REASON_PPP_START_FAILED = 12
NM_DEVICE_STATE_REASON_PPP_DISCONNECT = 13
NM_DEVICE_STATE_REASON_PPP_FAILED = 14
NM_DEVICE_STATE_REASON_DHCP_START_FAILED = 15
NM_DEVICE_STATE_REASON_DHCP_ERROR = 16
NM_DEVICE_STATE_REASON_DHCP_FAILED = 17
NM_DEVICE_STATE_REASON_SHARED_START_FAILED = 18
NM_DEVICE_STATE_REASON_SHARED_FAILED = 19
NM_DEVICE_STATE_REASON_AUTOIP_START_FAILED = 20
NM_DEVICE_STATE_REASON_AUTOIP_ERROR = 21
NM_DEVICE_STATE_REASON_AUTOIP_FAILED = 22
NM_DEVICE_STATE_REASON_MODEM_BUSY = 23
NM_DEVICE_STATE_REASON_MODEM_NO_DIAL_TONE = 24
NM_DEVICE_STATE_REASON_MODEM_NO_CARRIER = 25
NM_DEVICE_STATE_REASON_MODEM_DIAL_TIMEOUT = 26
NM_DEVICE_STATE_REASON_MODEM_DIAL_FAILED = 27
NM_DEVICE_STATE_REASON_MODEM_INIT_FAILED = 28
NM_DEVICE_STATE_REASON_GSM_APN_FAILED = 29
NM_DEVICE_STATE_REASON_GSM_REGISTRATION_NOT_SEARCHING = 30
NM_DEVICE_STATE_REASON_GSM_REGISTRATION_DENIED = 31
NM_DEVICE_STATE_REASON_GSM_REGISTRATION_TIMEOUT = 32
NM_DEVICE_STATE_REASON_GSM_REGISTRATION_FAILED = 33
NM_DEVICE_STATE_REASON_GSM_PIN_CHECK_FAILED = 34
NM_DEVICE_STATE_REASON_FIRMWARE_MISSING = 35
NM_DEVICE_STATE_REASON_REMOVED = 36
NM_DEVICE_STATE_REASON_SLEEPING = 37
NM_DEVICE_STATE_REASON_CONNECTION_REMOVED = 38
NM_DEVICE_STATE_REASON_USER_REQUESTED = 39
NM_DEVICE_STATE_REASON_CARRIER = 40
NM_DEVICE_STATE_REASON_CONNECTION_ASSUMED = 41
NM_DEVICE_STATE_REASON_SUPPLICANT_AVAILABLE = 42
NM_DEVICE_STATE_REASON_MODEM_NOT_FOUND = 43
NM_DEVICE_STATE_REASON_BT_FAILED = 44
NM_DEVICE_STATE_REASON_LAST = 0xFFFF

NM_802_11_AP_FLAGS_NONE = 0x00000000
NM_802_11_AP_FLAGS_PRIVACY = 0x00000001

NM_802_11_AP_SEC_NONE = 0x0
NM_802_11_AP_SEC_PAIR_WEP40 = 0x1
NM_802_11_AP_SEC_PAIR_WEP104 = 0x2
NM_802_11_AP_SEC_PAIR_TKIP = 0x4
NM_802_11_AP_SEC_PAIR_CCMP = 0x8
NM_802_11_AP_SEC_GROUP_WEP40 = 0x10
NM_802_11_AP_SEC_GROUP_WEP104 = 0x20
NM_802_11_AP_SEC_GROUP_TKIP = 0x40
NM_802_11_AP_SEC_GROUP_CCMP = 0x80
NM_802_11_AP_SEC_KEY_MGMT_PSK = 0x100
NM_802_11_AP_SEC_KEY_MGMT_802_1X = 0x200

NM_802_11_MODE_UNKNOWN = 0
NM_802_11_MODE_ADHOC = 1
NM_802_11_MODE_INFRA = 2

NM_WIFI_DEVICE_CAP_NONE = 0x00000000
NM_WIFI_DEVICE_CAP_CIPHER_WEP40 = 0x00000001
NM_WIFI_DEVICE_CAP_CIPHER_WEP104 = 0x00000002
NM_WIFI_DEVICE_CAP_CIPHER_TKIP = 0x00000004
NM_WIFI_DEVICE_CAP_CIPHER_CCMP = 0x00000008
NM_WIFI_DEVICE_CAP_WPA = 0x00000010
NM_WIFI_DEVICE_CAP_RSN = 0x00000020

NM_BT_CAPABILITY_NONE = 0x00000000
NM_BT_CAPABILITY_DUN = 0x00000001
NM_BT_CAPABILITY_NAP = 0x00000002

NM_DEVICE_MODEM_CAPABILITY_NONE = 0x00000000
NM_DEVICE_MODEM_CAPABILITY_POTS = 0x00000001
NM_DEVICE_MODEM_CAPABILITY_CDMA_EVDO = 0x00000002
NM_DEVICE_MODEM_CAPABILITY_GSM_UMTS = 0x00000004
NM_DEVICE_MODEM_CAPABILITY_LTE = 0x00000008

SETTINGS_SERVICE = 'org.freedesktop.NetworkManager'

NM_SERVICE = 'org.freedesktop.NetworkManager'
NM_IFACE = 'org.freedesktop.NetworkManager'
NM_PATH = '/org/freedesktop/NetworkManager'
NM_DEVICE_IFACE = 'org.freedesktop.NetworkManager.Device'
NM_WIRED_IFACE = 'org.freedesktop.NetworkManager.Device.Wired'
NM_WIRELESS_IFACE = 'org.freedesktop.NetworkManager.Device.Wireless'
NM_MODEM_IFACE = 'org.freedesktop.NetworkManager.Device.Modem'
NM_OLPC_MESH_IFACE = 'org.freedesktop.NetworkManager.Device.OlpcMesh'
NM_SETTINGS_PATH = '/org/freedesktop/NetworkManager/Settings'
NM_SETTINGS_IFACE = 'org.freedesktop.NetworkManager.Settings'
NM_CONNECTION_IFACE = 'org.freedesktop.NetworkManager.Settings.Connection'
NM_ACCESSPOINT_IFACE = 'org.freedesktop.NetworkManager.AccessPoint'
NM_ACTIVE_CONN_IFACE = 'org.freedesktop.NetworkManager.Connection.Active'

NM_SECRET_AGENT_IFACE = 'org.freedesktop.NetworkManager.SecretAgent'
NM_SECRET_AGENT_PATH = '/org/freedesktop/NetworkManager/SecretAgent'
NM_AGENT_MANAGER_IFACE = 'org.freedesktop.NetworkManager.AgentManager'
NM_AGENT_MANAGER_PATH = '/org/freedesktop/NetworkManager/AgentManager'

NM_AGENT_MANAGER_ERR_NO_SECRETS = \
    'org.freedesktop.NetworkManager.AgentManager.NoSecrets'

GSM_CONNECTION_ID = 'Sugar Modem Connection'
GSM_BAUD_RATE = 115200
GSM_KEYS_PATH = 'org.sugarlabs.network.gsm'
GSM_USERNAME_KEY = 'username'
GSM_PASSWORD_KEY = 'password'
GSM_NUMBER_KEY = 'number'
GSM_APN_KEY = 'apn'
GSM_PIN_KEY = 'pin'
GSM_PUK_KEY = 'puk'

ADHOC_CONNECTION_ID_PREFIX = 'Sugar Ad-hoc Network '
MESH_CONNECTION_ID_PREFIX = 'OLPC Mesh Network '
XS_MESH_CONNECTION_ID_PREFIX = 'OLPC XS Mesh Network '

_network_manager = None
_nm_settings = None
_secret_agent = None
_connections = None
_interfaces = None

_nm_device_state_reason_description = None


def get_error_by_reason(reason):
    global _nm_device_state_reason_description

    if _nm_device_state_reason_description is None:
        _nm_device_state_reason_description = {
            NM_DEVICE_STATE_REASON_UNKNOWN:
            _('The reason for the device state change is unknown.'),
            NM_DEVICE_STATE_REASON_NONE:
            _('The state change is normal.'),
            NM_DEVICE_STATE_REASON_NOW_MANAGED:
            _('The device is now managed.'),
            NM_DEVICE_STATE_REASON_NOW_UNMANAGED:
            _('The device is no longer managed.'),
            NM_DEVICE_STATE_REASON_CONFIG_FAILED:
            _('The device could not be readied for configuration.'),
            NM_DEVICE_STATE_REASON_IP_CONFIG_UNAVAILABLE:
            _('IP configuration could not be reserved '
              '(no available address, timeout, etc).'),
            NM_DEVICE_STATE_REASON_IP_CONFIG_EXPIRED:
            _('The IP configuration is no longer valid.'),
            NM_DEVICE_STATE_REASON_NO_SECRETS:
            _('Secrets were required, but not provided.'),
            NM_DEVICE_STATE_REASON_SUPPLICANT_DISCONNECT:
            _('The 802.1X supplicant disconnected from '
              'the access point or authentication server.'),
            NM_DEVICE_STATE_REASON_SUPPLICANT_CONFIG_FAILED:
            _('Configuration of the 802.1X supplicant failed.'),
            NM_DEVICE_STATE_REASON_SUPPLICANT_FAILED:
            _('The 802.1X supplicant quit or failed unexpectedly.'),
            NM_DEVICE_STATE_REASON_SUPPLICANT_TIMEOUT:
            _('The 802.1X supplicant took too long to authenticate.'),
            NM_DEVICE_STATE_REASON_PPP_START_FAILED:
            _('The PPP service failed to start within the allowed time.'),
            NM_DEVICE_STATE_REASON_PPP_DISCONNECT:
            _('The PPP service disconnected unexpectedly.'),
            NM_DEVICE_STATE_REASON_PPP_FAILED:
            _('The PPP service quit or failed unexpectedly.'),
            NM_DEVICE_STATE_REASON_DHCP_START_FAILED:
            _('The DHCP service failed to start within the allowed time.'),
            NM_DEVICE_STATE_REASON_DHCP_ERROR:
            _('The DHCP service reported an unexpected error.'),
            NM_DEVICE_STATE_REASON_DHCP_FAILED:
            _('The DHCP service quit or failed unexpectedly.'),
            NM_DEVICE_STATE_REASON_SHARED_START_FAILED:
            _('The shared connection service failed to start.'),
            NM_DEVICE_STATE_REASON_SHARED_FAILED:
            _('The shared connection service quit or failed'
              ' unexpectedly.'),
            NM_DEVICE_STATE_REASON_AUTOIP_START_FAILED:
            _('The AutoIP service failed to start.'),
            NM_DEVICE_STATE_REASON_AUTOIP_ERROR:
            _('The AutoIP service reported an unexpected error.'),
            NM_DEVICE_STATE_REASON_AUTOIP_FAILED:
            _('The AutoIP service quit or failed unexpectedly.'),
            NM_DEVICE_STATE_REASON_MODEM_BUSY:
            _('Dialing failed because the line was busy.'),
            NM_DEVICE_STATE_REASON_MODEM_NO_DIAL_TONE:
            _('Dialing failed because there was no dial tone.'),
            NM_DEVICE_STATE_REASON_MODEM_NO_CARRIER:
            _('Dialing failed because there was no carrier.'),
            NM_DEVICE_STATE_REASON_MODEM_DIAL_TIMEOUT:
            _('Dialing timed out.'),
            NM_DEVICE_STATE_REASON_MODEM_DIAL_FAILED:
            _('Dialing failed.'),
            NM_DEVICE_STATE_REASON_MODEM_INIT_FAILED:
            _('Modem initialization failed.'),
            NM_DEVICE_STATE_REASON_GSM_APN_FAILED:
            _('Failed to select the specified GSM APN'),
            NM_DEVICE_STATE_REASON_GSM_REGISTRATION_NOT_SEARCHING:
            _('Not searching for networks.'),
            NM_DEVICE_STATE_REASON_GSM_REGISTRATION_DENIED:
            _('Network registration was denied.'),
            NM_DEVICE_STATE_REASON_GSM_REGISTRATION_TIMEOUT:
            _('Network registration timed out.'),
            NM_DEVICE_STATE_REASON_GSM_REGISTRATION_FAILED:
            _('Failed to register with the requested GSM network.'),
            NM_DEVICE_STATE_REASON_GSM_PIN_CHECK_FAILED:
            _('PIN check failed.'),
            NM_DEVICE_STATE_REASON_FIRMWARE_MISSING:
            _('Necessary firmware for the device may be missing.'),
            NM_DEVICE_STATE_REASON_REMOVED:
            _('The device was removed.'),
            NM_DEVICE_STATE_REASON_SLEEPING:
            _('NetworkManager went to sleep.'),
            NM_DEVICE_STATE_REASON_CONNECTION_REMOVED:
            _("The device's active connection was removed "
              "or disappeared."),
            NM_DEVICE_STATE_REASON_USER_REQUESTED:
            _('A user or client requested the disconnection.'),
            NM_DEVICE_STATE_REASON_CARRIER:
            _("The device's carrier/link changed."),
            NM_DEVICE_STATE_REASON_CONNECTION_ASSUMED:
            _("The device's existing connection was assumed."),
            NM_DEVICE_STATE_REASON_SUPPLICANT_AVAILABLE:
            _("The supplicant is now available."),
            NM_DEVICE_STATE_REASON_MODEM_NOT_FOUND:
            _("The modem could not be found."),
            NM_DEVICE_STATE_REASON_BT_FAILED:
            _("The Bluetooth connection failed or timed out."),
            NM_DEVICE_STATE_REASON_LAST:
            _("Unused."),
        }

    return _nm_device_state_reason_description[reason]


def frequency_to_channel(frequency):
    """Returns the channel matching a given radio channel frequency. If a
    frequency is not in the dictionary channel 0 will be returned.

    Keyword arguments:
    frequency -- The radio channel frequency in MHz.

    Return: Channel represented by the frequency or 0

    """

    bg_table = {2412: 1, 2417: 2, 2422: 3, 2427: 4,
                2432: 5, 2437: 6, 2442: 7, 2447: 8,
                2452: 9, 2457: 10, 2462: 11, 2467: 12,
                2472: 13, 14: 2484}

    a_table = {5035: 7, 5040: 8, 5045: 9, 5055: 11,
               5060: 12, 5080: 16, 5170: 34,
               5180: 36, 5190: 38, 5200: 40,
               5210: 42, 5220: 44, 5230: 46,
               5240: 48, 5250: 50, 5260: 52,
               5280: 56, 5290: 58, 5300: 60,
               5320: 64, 5500: 100, 5520: 104,
               5540: 108, 5560: 112, 5580: 116,
               5600: 120, 5620: 124, 5640: 128,
               5660: 132, 5680: 136, 5700: 140,
               5745: 149, 5760: 152, 5765: 153,
               5785: 157, 5800: 160, 5805: 161,
               5825: 165, 4915: 183, 4920: 184,
               4925: 185, 4935: 187, 4945: 188,
               4960: 192, 4980: 196}
    if frequency not in bg_table and frequency not in a_table:
        logging.warning('The frequency %s can not be mapped to a channel, '
                        'returning 0.', frequency)
        return 0

    if frequency > 4900:
        return a_table[frequency]
    return bg_table[frequency]


def is_sugar_adhoc_network(ssid):
    """Checks whether an access point is a sugar Ad-hoc network.

    Keyword arguments:
    ssid -- Ssid of the access point.

    Return: Boolean

    """
    return ssid.startswith(b"Ad-hoc Network")


class WirelessSecurity(object):

    def __init__(self):
        self.key_mgmt = None
        self.proto = None
        self.group = None
        self.pairwise = None
        self.wep_key = None
        self.psk = None
        self.auth_alg = None

    def get_dict(self):
        wireless_security = {}
        if self.key_mgmt is not None:
            wireless_security['key-mgmt'] = self.key_mgmt
        if self.proto is not None:
            wireless_security['proto'] = self.proto
        if self.pairwise is not None:
            wireless_security['pairwise'] = self.pairwise
        if self.group is not None:
            wireless_security['group'] = self.group
        if self.wep_key is not None:
            wireless_security['wep-key0'] = self.wep_key
        if self.psk is not None:
            wireless_security['psk'] = self.psk
        if self.auth_alg is not None:
            wireless_security['auth-alg'] = self.auth_alg
        return wireless_security


class Wireless(object):
    nm_name = '802-11-wireless'

    def __init__(self):
        self.ssid = None
        self.security = None
        self.mode = None
        self.band = None
        self.channel = None

    def get_dict(self):
        wireless = {'ssid': self.ssid}
        if self.security:
            wireless['security'] = self.security
        if self.mode:
            wireless['mode'] = self.mode
        if self.band:
            wireless['band'] = self.band
        if self.channel:
            wireless['channel'] = self.channel
        return wireless


class OlpcMesh(object):
    nm_name = '802-11-olpc-mesh'

    def __init__(self, channel, anycast_addr):
        self.channel = channel
        self.anycast_addr = anycast_addr

    def get_dict(self):
        ret = {
            'ssid': dbus.ByteArray(b'olpc-mesh'),
            'channel': self.channel,
        }

        if self.anycast_addr:
            ret['dhcp-anycast-address'] = dbus.ByteArray(self.anycast_addr)
        return ret


class ConnectionSettings(object):

    def __init__(self):
        self.id = None
        self.uuid = None
        self.type = None
        self.autoconnect = False
        self.timestamp = None

    def get_dict(self):
        connection = {'id': self.id,
                      'uuid': self.uuid,
                      'type': self.type,
                      'autoconnect': self.autoconnect}
        if self.timestamp:
            connection['timestamp'] = self.timestamp
        return connection


class IP4Config(object):

    def __init__(self):
        self.method = None

    def get_dict(self):
        ip4_config = {}
        if self.method is not None:
            ip4_config['method'] = self.method
        return ip4_config


class Serial(object):

    def __init__(self):
        self.baud = None

    def get_dict(self):
        serial = {}

        if self.baud is not None:
            serial['baud'] = self.baud

        return serial


class Ppp(object):

    def __init__(self):
        pass

    def get_dict(self):
        ppp = {}
        return ppp


class Gsm(object):

    def __init__(self):
        self.apn = None
        self.number = None
        self.username = None
        self.pin = None
        self.password = None

    def get_dict(self):
        gsm = {}

        if self.apn:
            gsm['apn'] = self.apn
        if self.number:
            gsm['number'] = self.number
        if self.username:
            gsm['username'] = self.username
        if self.password:
            gsm['password'] = self.password
        if self.pin:
            gsm['pin'] = self.pin

        return gsm


class Settings(object):

    def __init__(self, wireless_cfg=None):
        self.connection = ConnectionSettings()
        self.ip4_config = None
        self.wireless_security = None

        if wireless_cfg is not None:
            self.wireless = wireless_cfg
        else:
            self.wireless = Wireless()

    def get_dict(self):
        settings = {}
        settings['connection'] = self.connection.get_dict()
        settings[self.wireless.nm_name] = self.wireless.get_dict()
        if self.wireless_security is not None:
            settings['802-11-wireless-security'] = \
                self.wireless_security.get_dict()
        if self.ip4_config is not None:
            settings['ipv4'] = self.ip4_config.get_dict()
        return settings


class SettingsGsm(object):

    def __init__(self):
        self.connection = ConnectionSettings()
        self.ip4_config = IP4Config()
        self.serial = Serial()
        self.ppp = Ppp()
        self.gsm = Gsm()

    def get_dict(self):
        settings = {}

        settings['connection'] = self.connection.get_dict()
        settings['serial'] = self.serial.get_dict()
        settings['ppp'] = self.ppp.get_dict()
        settings['gsm'] = self.gsm.get_dict()
        settings['ipv4'] = self.ip4_config.get_dict()

        return settings


class SecretsResponse(object):
    """Intermediate object to report the secrets from the dialog
    back to the connection object and which will inform NM
    """

    def __init__(self, reply_cb, error_cb):
        self._reply_cb = reply_cb
        self._error_cb = error_cb

    def set_secrets(self, secrets):
        self._reply_cb(secrets)

    def set_error(self, error):
        self._error_cb(error)


def set_connected():
    try:
        # try to flush resolver cache - SL#1940
        # ctypes' syntactic sugar does not work
        # so we must get the func ptr explicitly
        libc = ctypes.CDLL('libc.so.6')
        res_init = getattr(libc, '__res_init')
        res_init(None)
    except:
        # pylint: disable=W0702
        logging.exception('Error calling libc.__res_init')

    check_urgent_update()


class SecretAgent(dbus.service.Object):

    def __init__(self):
        self._bus = dbus.SystemBus()
        dbus.service.Object.__init__(self, self._bus, NM_SECRET_AGENT_PATH)
        self.secrets_request = dispatch.Signal()
        proxy = self._bus.get_object(NM_IFACE, NM_AGENT_MANAGER_PATH)
        proxy.Register("org.sugarlabs.sugar",
                       dbus_interface=NM_AGENT_MANAGER_IFACE,
                       reply_handler=self._register_reply_cb,
                       error_handler=self._register_error_cb)

    def _register_reply_cb(self):
        logging.debug("SecretAgent registered")

    def _register_error_cb(self, error):
        logging.error("Failed to register SecretAgent: %s", error)

    @dbus.service.method(NM_SECRET_AGENT_IFACE,
                         async_callbacks=('reply', 'error'),
                         in_signature='a{sa{sv}}osasb',
                         out_signature='a{sa{sv}}',
                         sender_keyword='sender',
                         byte_arrays=True)
    def GetSecrets(self, settings, connection_path, setting_name, hints,
                   request_new, reply, error, sender=None):
        if setting_name != '802-11-wireless-security':
            raise ValueError("Unsupported setting type %s" % (setting_name,))
        if not sender:
            raise Exception("Internal error: couldn't get sender")
        uid = self._bus.get_unix_user(sender)
        if uid != 0:
            raise Exception("UID %d not authorized" % (uid,))

        response = SecretsResponse(reply, error)
        self.secrets_request.send(self, settings=settings, response=response)


class AccessPoint(GObject.GObject):
    __gsignals__ = {
        'props-changed': (GObject.SignalFlags.RUN_FIRST, None,
                          ([GObject.TYPE_PYOBJECT])),
    }

    def __init__(self, device, model):
        GObject.GObject.__init__(self)
        self.device = device
        self.model = model

        self._initialized = False
        self._bus = dbus.SystemBus()

        self.ssid = b''
        self.strength = 0
        self.flags = 0
        self.wpa_flags = 0
        self.rsn_flags = 0
        self.mode = 0
        self.channel = 0

    def initialize(self):
        model_props = dbus.Interface(self.model, dbus.PROPERTIES_IFACE)
        model_props.GetAll(NM_ACCESSPOINT_IFACE, byte_arrays=True,
                           reply_handler=self._ap_properties_changed_cb,
                           error_handler=self._get_all_props_error_cb)

        self._bus.add_signal_receiver(self._ap_properties_changed_cb,
                                      signal_name='PropertiesChanged',
                                      path=self.model.object_path,
                                      dbus_interface=NM_ACCESSPOINT_IFACE,
                                      byte_arrays=True)

    def network_hash(self):
        """
        This is a hash which uniquely identifies the network that this AP
        is a bridge to. i.e. its expected for 2 APs with identical SSID and
        other settings to have the same network hash, because we assume that
        they are a part of the same underlying network.
        """

        # based on logic from nm-applet
        fl = 0

        if self.mode == NM_802_11_MODE_INFRA:
            fl |= 1 << 0
        elif self.mode == NM_802_11_MODE_ADHOC:
            fl |= 1 << 1
        else:
            fl |= 1 << 2

        # Separate out no encryption, WEP-only, and WPA-capable */
        if (not (self.flags & NM_802_11_AP_FLAGS_PRIVACY)) \
                and self.wpa_flags == NM_802_11_AP_SEC_NONE \
                and self.rsn_flags == NM_802_11_AP_SEC_NONE:
            fl |= 1 << 3
        elif (self.flags & NM_802_11_AP_FLAGS_PRIVACY) \
                and self.wpa_flags == NM_802_11_AP_SEC_NONE \
                and self.rsn_flags == NM_802_11_AP_SEC_NONE:
            fl |= 1 << 4
        elif (not (self.flags & NM_802_11_AP_FLAGS_PRIVACY)) \
                and self.wpa_flags != NM_802_11_AP_SEC_NONE \
                and self.rsn_flags != NM_802_11_AP_SEC_NONE:
            fl |= 1 << 5
        else:
            fl |= 1 << 6

        hashstr = str(fl) + '@' + self.ssid.decode()
        return hash(hashstr)

    def _update_properties(self, properties):
        if self._initialized:
            old_hash = self.network_hash()
        else:
            old_hash = None

        if 'Ssid' in properties:
            self.ssid = properties['Ssid']
        if 'Strength' in properties:
            self.strength = properties['Strength']
        if 'Flags' in properties:
            self.flags = properties['Flags']
        if 'WpaFlags' in properties:
            self.wpa_flags = properties['WpaFlags']
        if 'RsnFlags' in properties:
            self.rsn_flags = properties['RsnFlags']
        if 'Mode' in properties:
            self.mode = properties['Mode']
        if 'Frequency' in properties:
            self.channel = frequency_to_channel(properties['Frequency'])

        self._initialized = True
        self.emit('props-changed', old_hash)

    def _get_all_props_error_cb(self, err):
        logging.error('Error getting the access point properties: %s', err)

    def _ap_properties_changed_cb(self, properties):
        self._update_properties(properties)

    def disconnect(self):
        self._bus.remove_signal_receiver(self._ap_properties_changed_cb,
                                         signal_name='PropertiesChanged',
                                         path=self.model.object_path,
                                         dbus_interface=NM_ACCESSPOINT_IFACE)


def get_manager():
    global _network_manager
    if _network_manager is None:
        obj = dbus.SystemBus().get_object(NM_SERVICE, NM_PATH)
        _network_manager = dbus.Interface(obj, NM_IFACE)
    return _network_manager


def _get_settings():
    global _nm_settings
    if _nm_settings is None:
        obj = dbus.SystemBus().get_object(NM_SERVICE, NM_SETTINGS_PATH)
        _nm_settings = dbus.Interface(obj, NM_SETTINGS_IFACE)
        _migrate_old_wifi_connections()
        _migrate_old_gsm_connection()
    return _nm_settings


def get_secret_agent():
    global _secret_agent
    if _secret_agent is None:
        _secret_agent = SecretAgent()
    return _secret_agent


def _activate_reply_cb(connection_path):
    logging.debug('Activated connection: %s', connection_path)


def _activate_error_cb(err):
    logging.error('Failed to activate connection: %s', err)


def _add_and_activate_reply_cb(settings_path, connection_path):
    logging.debug('Added and activated connection: %s', connection_path)


def _add_and_activate_error_cb(err):
    logging.error('Failed to add and activate connection: %s', err)


class Connection(GObject.GObject):
    __gsignals__ = {
        'removed': (GObject.SignalFlags.RUN_LAST, None, ()),
    }

    def __init__(self, bus, path):
        GObject.GObject.__init__(self)
        obj = bus.get_object(NM_SERVICE, path)
        self._connection = dbus.Interface(obj, NM_CONNECTION_IFACE)
        self._removed_handle = self._connection.connect_to_signal(
            'Removed', self._removed_cb)
        self._updated_handle = self._connection.connect_to_signal(
            'Updated', self._updated_cb)
        self._settings = self._connection.GetSettings(byte_arrays=True)

    def _updated_cb(self):
        self._settings = self._connection.GetSettings(byte_arrays=True)

    def _removed_cb(self):
        self._updated_handle.remove()
        self._removed_handle.remove()
        self.emit('removed')

    def get_settings(self, stype=None):
        if not stype:
            return self._settings
        if stype in self._settings:
            return self._settings[stype]
        return None

    def get_secrets(self, stype, reply_handler, error_handler):
        return self._connection.GetSecrets(stype, byte_arrays=True,
                                           reply_handler=reply_handler,
                                           error_handler=error_handler)

    def update_settings(self, settings):
        self._connection.Update(settings)

    def activate(self, device_o, reply_handler=_activate_reply_cb,
                 error_handler=_activate_error_cb):
        activate_connection_by_path(self.get_path(), device_o,
                                    reply_handler=reply_handler,
                                    error_handler=error_handler)

    def delete(self):
        self._connection.Delete()

    def get_ssid(self):
        wifi_settings = self.get_settings('802-11-wireless')
        if wifi_settings and 'ssid' in wifi_settings:
            return wifi_settings['ssid']
        return None

    def get_id(self):
        return self.get_settings('connection')['id']

    def get_path(self):
        return self._connection.object_path


class Connections(object):

    def __init__(self):
        self._bus = dbus.SystemBus()
        self._connections = []

        settings = _get_settings()
        settings.connect_to_signal('NewConnection', self._new_connection_cb)

        for connection_o in settings.ListConnections():
            self._monitor_connection(connection_o)

    def get_list(self):
        return self._connections

    def _monitor_connection(self, connection_o):
        connection = Connection(self._bus, connection_o)
        connection.connect('removed', self._connection_removed_cb)
        self._connections.append(connection)

    def _new_connection_cb(self, connection_o):
        self._monitor_connection(connection_o)

    def _connection_removed_cb(self, connection):
        connection.disconnect_by_func(self._connection_removed_cb)
        self._connections.remove(connection)


def get_wireless_interfaces():
    global _interfaces
    if _interfaces is None:

        _interfaces = []
        bus = dbus.SystemBus()
        for device_path in get_manager().GetDevices():
            device_object = bus.get_object(NM_SERVICE, device_path)
            properties = dbus.Interface(device_object,
                                        'org.freedesktop.DBus.Properties')
            device_type = properties.Get(NM_DEVICE_IFACE, 'DeviceType')
            if device_type != NM_DEVICE_TYPE_WIFI:
                continue

            _interfaces.append(properties.Get(NM_DEVICE_IFACE, 'Interface'))
    return _interfaces


def get_connections():
    global _connections
    if _connections is None:
        _connections = Connections()
    return _connections


def find_connection_by_ssid(ssid):
    # FIXME: this check should be more extensive.
    # it should look at mode (infra/adhoc), band, security, and really
    # anything that is stored in the settings.
    for connection in get_connections().get_list():
        if connection.get_ssid() == ssid:
            return connection
    return None


def find_connection_by_id(connection_id):
    for connection in get_connections().get_list():
        if connection.get_id() == connection_id:
            return connection
    return None


def _add_connection_reply_cb(connection):
    logging.debug('Added connection: %s', connection)


def _add_connection_error_cb(err):
    logging.error('Failed to add connection: %s', err)


def add_connection(settings, reply_handler=_add_connection_reply_cb,
                   error_handler=_add_connection_error_cb):
    _get_settings().AddConnection(settings.get_dict(),
                                  reply_handler=reply_handler,
                                  error_handler=error_handler)


def activate_connection_by_path(connection, device_o,
                                reply_handler=_activate_reply_cb,
                                error_handler=_activate_error_cb):
    get_manager().ActivateConnection(connection,
                                     device_o,
                                     '/',
                                     reply_handler=reply_handler,
                                     error_handler=error_handler)


def add_and_activate_connection(device_o, settings, specific_object):
    manager = get_manager()
    manager.AddAndActivateConnection(settings.get_dict(), device_o,
                                     specific_object,
                                     reply_handler=_add_and_activate_reply_cb,
                                     error_handler=_add_and_activate_error_cb)


def _migrate_old_wifi_connections():
    """Migrate connections.cfg from Sugar-0.94 and previous to NetworkManager
    system-wide connections
    """

    profile_path = env.get_profile_path()
    config_path = os.path.join(profile_path, 'nm', 'connections.cfg')
    if not os.path.exists(config_path):
        return

    config = configparser.ConfigParser()
    try:
        if not config.read(config_path):
            logging.error('Error reading the nm config file')
            return
    except configparser.ParsingError:
        logging.exception('Error reading the nm config file')
        return

    for section in config.sections():
        try:
            settings = Settings()
            settings.connection.id = section
            ssid = config.get(section, 'ssid').encode()
            settings.wireless.ssid = dbus.ByteArray(ssid)
            config_uuid = config.get(section, 'uuid')
            settings.connection.uuid = config_uuid
            nmtype = config.get(section, 'type')
            settings.connection.type = nmtype
            autoconnect = bool(config.get(section, 'autoconnect'))
            settings.connection.autoconnect = autoconnect

            if config.has_option(section, 'timestamp'):
                timestamp = int(config.get(section, 'timestamp'))
                settings.connection.timestamp = timestamp

            if config.has_option(section, 'key-mgmt'):
                settings.wireless_security = WirelessSecurity()
                mgmt = config.get(section, 'key-mgmt')
                settings.wireless_security.key_mgmt = mgmt
                security = config.get(section, 'security')
                settings.wireless.security = security
                key = config.get(section, 'key')
                if mgmt == 'none':
                    settings.wireless_security.wep_key = key
                    auth_alg = config.get(section, 'auth-alg')
                    settings.wireless_security.auth_alg = auth_alg
                elif mgmt == 'wpa-psk':
                    settings.wireless_security.psk = key
                    if config.has_option(section, 'proto'):
                        value = config.get(section, 'proto')
                        settings.wireless_security.proto = value
                    if config.has_option(section, 'group'):
                        value = config.get(section, 'group')
                        settings.wireless_security.group = value
                    if config.has_option(section, 'pairwise'):
                        value = config.get(section, 'pairwise')
                        settings.wireless_security.pairwise = value
        except configparser.Error:
            logging.exception('Error reading section')
        else:
            add_connection(settings)

    os.unlink(config_path)


def create_gsm_connection(username, password, number, apn, pin):
    settings = SettingsGsm()
    settings.gsm.username = username
    settings.gsm.number = number
    settings.gsm.apn = apn
    settings.gsm.pin = pin
    settings.gsm.password = password

    settings.connection.id = GSM_CONNECTION_ID
    settings.connection.type = NM_CONNECTION_TYPE_GSM
    settings.connection.uuid = str(uuid.uuid4())
    settings.connection.autoconnect = False
    settings.ip4_config.method = 'auto'
    settings.serial.baud = GSM_BAUD_RATE

    add_connection(settings)


def _migrate_old_gsm_connection():
    if find_gsm_connection():
        # don't attempt migration if a NM-level connection already exists
        return

    settings = Gio.Settings.new(GSM_KEYS_PATH)

    username = settings.get_string(GSM_USERNAME_KEY) or ''
    password = settings.get_string(GSM_PASSWORD_KEY) or ''
    number = settings.get_string(GSM_NUMBER_KEY) or ''
    apn = settings.get_string(GSM_APN_KEY) or ''
    pin = settings.get_string(GSM_PIN_KEY) or ''

    if apn or number:
        logging.info("Migrating old GSM connection details")
        try:
            create_gsm_connection(username, password, number, apn, pin)
            # remove old connection
            for setting in (GSM_USERNAME_KEY, GSM_PASSWORD_KEY,
                            GSM_NUMBER_KEY, GSM_APN_KEY, GSM_PIN_KEY,
                            GSM_PUK_KEY):
                settings.set_string(setting, '')
        except Exception:
            logging.exception('Error adding gsm connection to NMSettings.')


def find_gsm_connection():
    return find_connection_by_id(GSM_CONNECTION_ID)


def disconnect_access_points(ap_paths):
    """
    Disconnect all devices connected to any of the given access points.
    """
    bus = dbus.SystemBus()
    netmgr_obj = bus.get_object(NM_SERVICE, NM_PATH)
    netmgr = dbus.Interface(netmgr_obj, NM_IFACE)
    netmgr_props = dbus.Interface(netmgr, dbus.PROPERTIES_IFACE)
    active_connection_paths = netmgr_props.Get(NM_IFACE, 'ActiveConnections')

    for conn_path in active_connection_paths:
        conn_obj = bus.get_object(NM_IFACE, conn_path)
        conn_props = dbus.Interface(conn_obj, dbus.PROPERTIES_IFACE)
        ap_path = conn_props.Get(NM_ACTIVE_CONN_IFACE, 'SpecificObject')
        if ap_path == '/' or ap_path not in ap_paths:
            continue

        dev_paths = conn_props.Get(NM_ACTIVE_CONN_IFACE, 'Devices')
        for dev_path in dev_paths:
            dev_obj = bus.get_object(NM_SERVICE, dev_path)
            dev = dbus.Interface(dev_obj, NM_DEVICE_IFACE)
            dev.Disconnect()


def forget_wireless_network(ssid):
    connection = find_connection_by_ssid(ssid)
    if connection:
        connection.delete()


def _is_non_printable(char):
    """
    Return True if char is a non-printable unicode character, False otherwise
    """
    return (char < ' ') or ('~' < char < '\xA0') or (char == '\xAD')


def ssid_to_display_name(ssid):
    """Convert an SSID into a unicode string for recognising Access Points

    Return a unicode string that's useful for recognising and
    distinguishing between Access Points (APs).

    IEEE 802.11 defines SSIDs as arbitrary byte sequences. As random
    bytes are not very user-friendly, most APs use some human-readable
    character string as SSID. However, because there's no standard
    specifying what encoding to use, AP vendors chose various
    different encodings. Since there's also no indication of what
    encoding was used for a particular SSID, the best we can do for
    turning an SSID into a displayable string is to try a couple of
    encodings based on some heuristic.

    We're currently using the following heuristic:

    1. If the SSID is a valid character string consisting only of
       printable characters in one of the following encodings (tried in
       the given order), decode it accordingly:
       UTF-8, ISO-8859-1, Windows-1251.
    2. Return a hex dump of the SSID.
    """
    for encoding in ['utf-8', 'iso-8859-1', 'windows-1251']:
        try:
            display_name = str(ssid, encoding)
        except UnicodeDecodeError:
            continue

        if not [True for char in display_name if _is_non_printable(char)]:
            # Only printable characters
            return display_name

    return ':'.join(['%02x' % (ord(byte), ) for byte in ssid])


wifi_whitelist = ('Sugar Ad-hoc Network 1',
                  'Sugar Ad-hoc Network 6',
                  'Sugar Ad-hoc Network 11')
mesh_whitelist = ('OLPC Mesh Network 1',
                  'OLPC Mesh Network 6',
                  'OLPC Mesh Network 11',
                  'OLPC XS Mesh Network 1',
                  'OLPC XS Mesh Network 6',
                  'OLPC XS Mesh Network 11')


def is_wireless(connection):
    """Check for wireless connection not whitelisted by Sugar.
    """
    wifi_settings = connection.get_settings(
        NM_CONNECTION_TYPE_802_11_WIRELESS)
    if wifi_settings:
        return not (wifi_settings['mode'] == 'adhoc' and
                    connection.get_id() in wifi_whitelist)

    mesh_settings = connection.get_settings(
        NM_CONNECTION_TYPE_802_11_OLPC_MESH)
    if mesh_settings:
        return not connection.get_id() in mesh_whitelist


def clear_wireless_networks():
    """Remove all wireless connections except Sugar-internal ones.
    """
    try:
        connections = get_connections()
    except dbus.DBusException:
        logging.debug('NetworkManager not available')
    else:
        wireless_connections = \
            (connection for connection in
             connections.get_list() if is_wireless(connection))

        for connection in wireless_connections:
            try:
                connection.delete()
            except dbus.DBusException:
                logging.debug("Could not remove connection %s",
                              connection.get_id())


def have_wireless_networks():
    """Check that there are non-Sugar-internal wireless connections.
    """
    try:
        connections = get_connections()
    except dbus.DBusException:
        logging.debug('NetworkManager not available')
        return False
    else:
        return any(is_wireless(connection)
                   for connection in connections.get_list())



# Copyright (C) 2007, 2008 One Laptop Per Child
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
# The language config is based on the system-config-language
# (http://fedoraproject.org/wiki/SystemConfig/language) tool
# Parts of the code were reused.
#

import os
import locale
from gettext import gettext as _
import subprocess

_default_lang = '%s.%s' % locale.getdefaultlocale()
_standard_msg = _('Could not access ~/.i18n. Create standard settings.')

LANGUAGES = {
    "Afar": "Afar",
    "Afrikaans": "Afrikaans",
    "Aguaruna": "Agr",
    "Akan": "Akan",
    "Albanian": "Shqip",
    "American English": "English",
    "Amharic": "\u12a0\u121b\u122d\u129b",
    "Angika": "Angika",
    "Arabic": "\u0627\u0644\u0639\u0631\u0628\u064a\u0629",
    "Aragonese": "Aragonese",
    "Armenian": "\u0540\u0561\u0575\u0565\u0580\u0565\u0576",
    "Assamese": "\u0985\u09b8\u09ae\u09c0\u09af\u09bc\u09be",
    "Asturian": "Asturianu",
    "Australian English": "English",
    "Austrian German": "Deutsch",
    "Aymara": "Ayc",
    "Azerbaijani": "Az\u0259rbaycan",
    "Bangla": "\u09ac\u09be\u0982\u09b2\u09be",
    "Basque": "Euskara",
    "Belarusian": "\u0411\u0435\u043b\u0430\u0440\u0443\u0441\u043a\u0430\u044f",
    "Bemba": "Ichibemba",
    "Berber": "Ber",
    "Bhili": "Bhb",
    "Bhojpuri": "Bhojpuri",
    "Bislama": "Bislama",
    "Blin": "Blin",
    "Bodo": "\u092c\u0921\u093c\u094b",
    "Bosnian": "Bosanski",
    "Brazilian Portuguese": "Portugu\u00eas",
    "Breton": "Brezhoneg",
    "British English": "English",
    "Bulgarian": "\u0411\u044a\u043b\u0433\u0430\u0440\u0441\u043a\u0438",
    "Burmese": "\u1019\u103c\u1014\u103a\u1019\u102c",
    "Canadian English": "English",
    "Canadian French": "Fran\u00e7ais",
    "Cantonese": "\u7cb5\u8a9e",
    "Catalan": "Catal\u00e0",
    "Central Kurdish": "\u06a9\u0648\u0631\u062f\u06cc\u06cc",
    "Central Nahuatl": "Nhn",
    "Chechen": "\u041d\u043e\u0445\u0447\u0438\u0439\u043d",
    "Cherokee": "\u13e3\uab83\uab79",
    "Chhattisgarhi": "Hne",
    "Chinese": "\u4e2d\u6587\uff08\u4e2d\u56fd\uff09",
    "Chitwania Tharu": "The",
    "Chuvash": "Chuvash",
    "Cornish": "Kernewek",
    "Crimean Tatar": "Crimean",
    "Croatian": "Hrvatski",
    "Cusco Quechua": "Quz",
    "Czech": "\u010ce\u0161tina",
    "Danish": "Dansk",
    "Divehi": "Divehi",
    "Dogri": "\u0921\u094b\u0917\u0930\u0940",
    "Dutch": "Nederlands",
    "Dzongkha": "\u0f62\u0fab\u0f7c\u0f44\u0f0b\u0f41\u0f0d",
    "English": "English",
    "Estonian": "Eesti",
    "European Portuguese": "Portugu\u00eas",
    "European Spanish": "Espa\u00f1ol",
    "Faroese": "F\u00f8royskt",
    "Fiji Hindi": "Fiji",
    "Filipino": "Filipino",
    "Finnish": "Suomi",
    "Flemish": "Nederlands",
    "French": "Fran\u00e7ais",
    "Friulian": "Furlan",
    "Fulah": "Pulaar",
    "Galician": "Galego",
    "Ganda": "Luganda",
    "Geez": "Geez",
    "Georgian": "\u10e5\u10d0\u10e0\u10d7\u10e3\u10da\u10d8",
    "German": "Deutsch",
    "Greek": "\u0395\u03bb\u03bb\u03b7\u03bd\u03b9\u03ba\u03ac",
    "Gujarati": "\u0a97\u0ac1\u0a9c\u0ab0\u0abe\u0aa4\u0ac0",
    "Haitian Creole": "Haitian",
    "Hakka Chinese": "Hakka",
    "Hausa": "Hausa",
    "Hebrew": "\u05e2\u05d1\u05e8\u05d9\u05ea",
    "Hindi": "\u0939\u093f\u0928\u094d\u0926\u0940",
    "Hungarian": "Magyar",
    "Icelandic": "\u00cdslenska",
    "Igbo": "Igbo",
    "Indonesian": "Indonesia",
    "Interlingua": "Interlingua",
    "Inuktitut": "Inuktitut",
    "Inupiaq": "Inupiaq",
    "Irish": "Gaeilge",
    "Italian": "Italiano",
    "Japanese": "\u65e5\u672c\u8a9e",
    "Kabyle": "Taqbaylit",
    "Kalaallisut": "Kalaallisut",
    "Kannada": "\u0c95\u0ca8\u0ccd\u0ca8\u0ca1",
    "Karbi": "Mjw",
    "Kashmiri": "\u06a9\u0672\u0634\u064f\u0631",
    "Kashubian": "Kashubian",
    "Kazakh": "\u049a\u0430\u0437\u0430\u049b",
    "Khmer": "\u1781\u17d2\u1798\u17c2\u179a",
    "Kinyarwanda": "Kinyarwanda",
    "Konkani": "\u0915\u094b\u0902\u0915\u0923\u0940",
    "Korean": "\ud55c\uad6d\uc5b4(\ub300\ud55c\ubbfc\uad6d)",
    "Kurdish": "Kurd\u00ee",
    "Kyrgyz": "\u041a\u044b\u0440\u0433\u044b\u0437\u0447\u0430",
    "Lao": "\u0ea5\u0eb2\u0ea7",
    "Latvian": "Latvie\u0161u",
    "Ligurian": "Ligurian",
    "Limburgish": "Limburgish",
    "Lingala": "Ling\u00e1la",
    "Literary Chinese": "Literary",
    "Lithuanian": "Lietuvi\u0173",
    "Low German": "Low",
    "Low Saxon": "Low",
    "Lower Sorbian": "Dolnoserb\u0161\u0107ina",
    "Luxembourgish": "L\u00ebtzebuergesch",
    "Macedonian": "\u041c\u0430\u043a\u0435\u0434\u043e\u043d\u0441\u043a\u0438",
    "Magahi": "Magahi",
    "Maithili": "\u092e\u0948\u0925\u093f\u0932\u0940",
    "Malagasy": "Malagasy",
    "Malay": "Melayu",
    "Malayalam": "\u0d2e\u0d32\u0d2f\u0d3e\u0d33\u0d02",
    "Maltese": "Malti",
    "Mandarin Chinese": "Cmn",
    "Manipuri": "\u09ae\u09c8\u09a4\u09c8\u09b2\u09cb\u09a8\u09cd",
    "Manx": "Gaelg",
    "Maori": "Te",
    "Marathi": "\u092e\u0930\u093e\u0920\u0940",
    "Meadow Mari": "Mhr",
    "Mexican Spanish": "Espa\u00f1ol",
    "Min Nan Chinese": "Min",
    "Miskito": "Miq",
    "Mon": "Mnw",
    "Mongolian": "\u041c\u043e\u043d\u0433\u043e\u043b",
    "Morisyen": "Kreol",
    "Nepali": "\u0928\u0947\u092a\u093e\u0932\u0940",
    "Niuean": "Niuean",
    "Northern Sami": "Davvis\u00e1megiella",
    "Northern Sotho": "Northern",
    "Norwegian Bokm\u00e5l": "Norsk",
    "Norwegian Nynorsk": "Norsk",
    "Occitan": "Occitan",
    "Odia": "\u0b13\u0b21\u0b3c\u0b3f\u0b06",
    "Oromo": "Oromoo",
    "Ossetic": "\u0418\u0440\u043e\u043d",
    "Papiamento": "Papiamento",
    "Pashto": "\u067e\u069a\u062a\u0648",
    "Persian": "\u0641\u0627\u0631\u0633\u06cc",
    "Polish": "Polski",
    "Punjabi": "\u0a2a\u0a70\u0a1c\u0a3e\u0a2c\u0a40",
    "Rajasthani": "Rajasthani",
    "Romanian": "Rom\u00e2n\u0103",
    "Russian": "\u0420\u0443\u0441\u0441\u043a\u0438\u0439",
    "Sakha": "\u0421\u0430\u0445\u0430",
    "Samoan": "Samoan",
    "Samogitian": "Samogitian",
    "Sanskrit": "\u0938\u0902\u0938\u094d\u0915\u0943\u0924",
    "Santali": "\u1c65\u1c5f\u1c71\u1c5b\u1c5f\u1c72\u1c64",
    "Sardinian": "Sardinian",
    "Scottish Gaelic": "G\u00e0idhlig",
    "Serbian": "\u0421\u0440\u043f\u0441\u043a\u0438",
    "Shan": "Shan",
    "Shuswap": "Shs",
    "Sidamo": "Sidamo",
    "Silesian": "Silesian",
    "Sindhi": "\u0633\u0646\u068c\u064a",
    "Sinhala": "\u0dc3\u0dd2\u0d82\u0dc4\u0dbd",
    "Slovak": "Sloven\u010dina",
    "Slovenian": "Sloven\u0161\u010dina",
    "Somali": "Soomaali",
    "South Azerbaijani": "Az\u0259rbaycan",
    "South Ndebele": "South",
    "Southern Sotho": "Southern",
    "Spanish": "Espa\u00f1ol",
    "Swahili": "Kiswahili",
    "Swati": "Swati",
    "Swedish": "Svenska",
    "Swiss French": "Fran\u00e7ais",
    "Swiss High German": "Deutsch",
    "Tagalog": "Tagalog",
    "Tajik": "\u0422\u043e\u04b7\u0438\u043a\u04e3",
    "Tamil": "\u0ba4\u0bae\u0bbf\u0bb4\u0bcd",
    "Tatar": "\u0422\u0430\u0442\u0430\u0440",
    "Telugu": "\u0c24\u0c46\u0c32\u0c41\u0c17\u0c41",
    "Thai": "\u0e44\u0e17\u0e22",
    "Tibetan": "\u0f56\u0f7c\u0f51\u0f0b\u0f66\u0f90\u0f51\u0f0b",
    "Tigre": "Tigre",
    "Tigrinya": "\u1275\u130d\u122d",
    "Tok Pisin": "Tok",
    "Tongan": "Lea",
    "Tsonga": "Tsonga",
    "Tswana": "Tswana",
    "Tulu": "Tulu",
    "Turkish": "T\u00fcrk\u00e7e",
    "Turkmen": "T\u00fcrkmen",
    "Ukrainian": "\u0423\u043a\u0440\u0430\u0457\u043d\u0441\u044c\u043a\u0430",
    "Unami Delaware": "Unm",
    "Upper Sorbian": "Hornjoserb\u0161\u0107ina",
    "Urdu": "\u0627\u0631\u062f\u0648",
    "Uyghur": "\u0626\u06c7\u064a\u063a\u06c7\u0631\u0686\u06d5",
    "Uzbek": "O\u2018Zbek",
    "Venda": "Venda",
    "Vietnamese": "Ti\u1ebfng",
    "Walloon": "Walloon",
    "Walser": "Walser",
    "Welsh": "Cymraeg",
    "Western Frisian": "Frysk",
    "Wolaytta": "Wolaytta",
    "Wolof": "Wolof",
    "Xhosa": "Isixhosa",
    "YauNungon": "Yuw",
    "Yiddish": "\u05d9\u05d9\u05b4\u05d3\u05d9\u05e9",
    "Yoruba": "\u00c8d\u00e8",
    "Zulu": "Isizulu"
}



def read_all_languages():
    fdp = subprocess.Popen(['locale', '-av'], stdout=subprocess.PIPE)
    lines = fdp.stdout.read().decode('utf-8', 'ignore').split('\n')
    locales = []

    for line in lines:
        if line.find('locale:') != -1:
            locale_str = line.split()[1]
        elif line.find('title |') != -1:
            title = line.lstrip('title |')
        elif line.find('language |') != -1:
            lang = line.lstrip('language |')
            # Sometimes language is a language code, not the language name
            if len(lang) <= 3:
                lang = title.split()[0]
        elif line.find('territory |') != -1:
            territory = line.lstrip('territory |')
            # Sometimes territory is a territory code, not the territory name
            if len(territory) <= 3 and territory != 'USA':
                if ' locale for ' in title:
                    territory = title.split(' locale for ')[-1]
                    # Aesthetic cleanup up for titles with trailing .
                    if territory[-1] == '.':
                        territory = territory[:-1]
                else:
                    territory = title.split()[-1]
            if locale_str.endswith('utf8') and len(lang):
                locales.append((LANGUAGES[lang], territory, locale_str))

    locales.sort()
    return locales


def _initialize():
    if set_languages.__doc__ is None:
        # when running under 'python -OO', all __doc__ fields are None,
        # so += would fail -- and this function would be unnecessary anyway.
        return
    languages = read_all_languages()
    set_languages.__doc__ += '\n'
    for lang in languages:
        set_languages.__doc__ += '%s \n' % (lang[0].replace(' ', '_') + '/' +
                                            lang[1].replace(' ', '_'))


def _write_i18n(lang_env, language_env):
    path = os.path.join(os.environ.get('HOME', ''), '.i18n')
    # try avoid writing the file if the values didn't changed
    lang_line = None
    language_line = None
    other_lines = []
    try:
        with open(path) as fd:
            for line in fd:
                if line.startswith('LANG='):
                    lang_line = line
                elif line.startswith('LANGUAGE='):
                    language_line = line
                else:
                    other_lines.append(line)
    except:
        pass

    new_lang_line = 'LANG="%s"\n' % lang_env
    new_language_line = 'LANGUAGE="%s"\n' % language_env

    if lang_line == new_lang_line and language_line == new_language_line:
        return

    with open(path, 'w') as fd:
        fd.write(new_lang_line)
        fd.write(new_language_line)
        for line in other_lines:
            fd.write(line)
        fd.flush()
        # be sure all is flushed
        os.fsync(fd)


def _get_from_env(name):
    lang = os.environ[name]
    lang = lang.strip()
    if lang.endswith('UTF-8'):
        lang = lang.replace('UTF-8', 'utf8')
    return lang


def get_languages():
    # read the env variables set in bin/sugar
    langlist = None
    lang = _default_lang

    if 'LANGUAGE' in os.environ:
        lang = _get_from_env('LANGUAGE')
        langlist = lang.split(':')

    if 'LANG' in os.environ:
        lang = _get_from_env('LANG')

    # There might be cases where .i18n may not contain a LANGUAGE field
    if langlist is None:
        return [lang]
    return langlist


def print_languages():
    codes = get_languages()

    languages = read_all_languages()
    for code in codes:
        found_lang = False
        for lang in languages:
            if lang[2].split('.')[0] == code.split('.')[0]:
                print(lang[0].replace(' ', '_') + '/' +
                      lang[1].replace(' ', '_'))
                found_lang = True
                break
        if not found_lang:
            print((_('Language for code=%s could not be determined.') % code))


def set_languages(languages):
    """Set the system language.
    languages :
    """

    if isinstance(languages, list):
        set_languages_list(languages)
        return

    if languages.endswith('utf8'):
        set_languages_list([languages])
        return 1
    langs = read_all_languages()
    for lang, territory, locale_str in langs:
        code = lang.replace(' ', '_') + '/' \
                    + territory.replace(' ', '_')
        if code == languages:
            set_languages_list([locale_str])
            return 1
    print((_("Sorry I do not speak \'%s\'.") % languages))


def set_languages_list(languages):
    """Set the system language using a list of preferred languages"""
    colon = ':'
    language_env = colon.join(languages)
    lang_env = languages[0].strip('\n')
    _write_i18n(lang_env, language_env)


# inilialize the docstrings for the language
_initialize()



# Copyright (C) 2008 One Laptop Per Child
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import logging

from sugar3.graphics import tray

from jarabe import config


class DevicesTray(tray.HTray):

    def __init__(self):
        tray.HTray.__init__(self, align=tray.ALIGN_TO_END)

        for f in os.listdir(os.path.join(config.ext_path, 'deviceicon')):
            if f.endswith('.py') and not f.startswith('__'):
                module_name = f[:-3]
                try:
                    mod = __import__('deviceicon.' + module_name, globals(),
                                     locals(), [module_name])
                    mod.setup(self)
                except Exception:
                    logging.exception('Exception while loading extension:')

    def add_device(self, view):
        index = 0
        relative_index = getattr(view, 'FRAME_POSITION_RELATIVE', -1)
        for item in self.get_children():
            current_relative_index = getattr(item, 'FRAME_POSITION_RELATIVE',
                                             0)
            if current_relative_index >= relative_index:
                index += 1
            else:
                break
        self.add_item(view, index=index)
        view.show()

    def remove_device(self, view):
        self.remove_item(view)
